from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.core.logic_gate_circuits import (
    MultipleInputLogicGate
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder,
    TwoOneMultiplexer
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    XorGate
)


class UnsignedCarrySkipAdder(GeneralCircuit):
    """Class representing unsigned carry skip (bypass) adder composed of smaller carry bypass blocks of chosen size to reduce propagation delay.

    Unsigned carry skip (bypass) adder represents faster adder circuit which is composed
    of more complex circuitry but provides much less propagation delay as opposed to rca.

    Each carry bypass block is composed of these logic parts:
    Propagate XOR gates compute propagate signals of corresponding bit pairs, these signals
    are then combined in multiple input AND gate (cascaded two input gates).
    Half/full adder cascade represents basic ripple carry adder design for input carry to ripple through them,
    additionally these adders compute individual output sum bits.
    Finally multiplexer lies at the end of each carry bypass block and is used to propagate block's input carry
    if multiple input AND gate output, which serves as select signal, is 1 or to wait for rippling of cout from the block's adders if it is 0.

    ```
                     ┼   ┼                      ┼   ┼
                 ┌───▼───▼───┐              ┌───▼───▼───┐
            ┌────┤ Propagate │         ┌────┤ Propagate │
            │SEL │  signals  │         │SEL │  signals  │
       ┌────▼─┐  └───────────┘    ┌────▼─┐  └───────────┘
       │      │                   │      │
    ┌──┤2:1MUX│◄────────────────┬─┤2:1MUX│◄────────────────┬─Cin
    │  │      │                 │ │      │                 │
    │  └────▲─┘      ┼   ┼      │ └────▲─┘      ┼   ┼      │
    │       │    ┌───▼───▼───┐  │      │    ┌───▼───▼───┐  │
    │       │    │  Adders   │◄─┘      │    │  Adders   │◄─┘
    │       └────┤           │         └────┤           │
    │            └─────┬─────┘              └─────┬─────┘
    │                  ┼                          ┼
    ▼                  ▼                          ▼
    Cout               Sums                       Sums
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        bypass_block_size (int, optional): Size of each composite bypass adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of unsigned cska. Defaults to "".
        name (str, optional): Name of unsigned cska. Defaults to "u_cska".
    """
    def __init__(self, a: Bus, b: Bus, bypass_block_size: int = 4, prefix: str = "", name: str = "u_cska", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        assert bypass_block_size > 0, "Block size of the individual bypass groups must be greater than 0."

        # To signify current number of blocks and number of bits that remain to be added into function blocks
        block_n = 0
        N_wires = self.N
        cin = ConstantWireValue0()
        while N_wires != 0:
            propagate_wires = []
            block_size = bypass_block_size if N_wires >= bypass_block_size else N_wires
            for i in range(block_size):
                # Generate propagate wires for corresponding bit pairs
                propagate_xor = XorGate(a=self.a.get_wire((block_n*bypass_block_size)+i), b=self.b.get_wire((block_n*bypass_block_size)+i), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
                self.add_component(propagate_xor)
                propagate_wires.append(propagate_xor.out)

                if block_n == 0 and i == 0:
                    obj_adder = HalfAdder(a=self.a.get_wire((block_n*bypass_block_size)+i), b=self.b.get_wire((block_n*bypass_block_size)+i), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                else:
                    obj_adder = FullAdder(a=self.a.get_wire((block_n*bypass_block_size)+i), b=self.b.get_wire((block_n*bypass_block_size)+i), c=cout, prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))

                cout = obj_adder.get_carry_wire()
                self.add_component(obj_adder)
                # Connecting adder's output sum bit to its proper position within the described circuit's output bus
                self.out.connect(i+(block_n*bypass_block_size), obj_adder.get_sum_wire())

            # ANDing of propagate wires, gate's output serves as select signal into 2:1 multiplexer and signifies whether block's input carry should be propagated (thus reducing delay) or not
            propagation_and = MultipleInputLogicGate(a=Bus(prefix=self.prefix+f"_propagate_signals{block_n}", N=len(propagate_wires), wires_list=propagate_wires), two_input_gate_cls=AndGate, parent_component=self, prefix=self.prefix+f"_and_propagate{block_n}")

            mux = TwoOneMultiplexer(a=cout, b=cin, c=propagation_and.out, prefix=self.prefix+"_mux2to1"+str(self.get_instance_num(cls=TwoOneMultiplexer)))
            self.add_component(mux)

            # Updating cin for the the next bypass block
            # Also updating cout value which is used as cin for the first adder of the next block
            cin = mux.out.get_wire()
            cout = mux.out.get_wire()
            N_wires -= block_size
            block_n += 1

        # Connection of final Cout
        self.out.connect(self.N, cin)


class SignedCarrySkipAdder(UnsignedCarrySkipAdder, GeneralCircuit):
    """Class representing signed carry skip (bypass) adder composed of smaller carry bypass blocks of chosen size to reduce propagation delay.

    Signed carry skip (bypass) adder represents faster adder circuit which is composed
    of more complex circuitry but provides much less propagation delay as opposed to rca.

    Each carry bypass block is composed of these logic parts:
    Propagate XOR gates compute propagate signals of corresponding bit pairs, these signals
    are then combined in multiple input AND gate (cascaded two input gates).
    Half/full adder cascade represents basic ripple carry adder design for input carry to ripple through them,
    additionally these adders compute individual output sum bits.
    Finally multiplexer lies at the end of each carry bypass block and is used to propagate block's input carry
    if multiple input AND gate output, which serves as select signal, is 1 or to wait for rippling of cout from the block's adders if it is 0.

    At last XOR gates are used to ensure proper sign extension.

    ```
                       ┼   ┼                      ┼   ┼
                   ┌───▼───▼───┐              ┌───▼───▼───┐
              ┌────┤ Propagate │         ┌────┤ Propagate │
              │SEL │  signals  │         │SEL │  signals  │
         ┌────▼─┐  └───────────┘    ┌────▼─┐  └───────────┘
         │      │                   │      │
       ┌─┤2:1MUX│◄────────────────┬─┤2:1MUX│◄────────────────┬─Cin
       │ │      │                 │ │      │                 │
       │ └────▲─┘      ┼   ┼      │ └────▲─┘      ┼   ┼      │
    ┌──▼───┐  │    ┌───▼───▼───┐  │      │    ┌───▼───▼───┐  │
    │ SIGN │  │    │  Adders   │◄─┘      │    │  Adders   │◄─┘
    │Extend│  └────┤           │         └────┤           │
    └──┬───┘       └─────┬─────┘              └─────┬─────┘
       │                 ┼                          ┼
       ▼                 ▼                          ▼
      Cout              Sums                       Sums
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        bypass_block_size (int, optional): Size of each composite bypass adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of signed cska. Defaults to "".
        name (str, optional): Name of signed cska. Defaults to "s_cska".
    """
    def __init__(self, a: Bus, b: Bus, bypass_block_size: int = 4, prefix: str = "", name: str = "s_cska", **kwargs):
        # Signed bus length extension
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))
        
        super().__init__(a=a, b=b, bypass_block_size=bypass_block_size, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        sign_xor_1 = XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self)
        self.add_component(sign_xor_1)
        sign_xor_2 = XorGate(sign_xor_1.out, self.get_previous_component(2).out.get_wire(), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self)
        self.add_component(sign_xor_2)
        self.out.connect(self.N, sign_xor_2.out)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    FullAdder,
    TwoOneMultiplexer
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    OrGate,
    XorGate
)


class UnsignedCarrySelectAdder(GeneralCircuit):
    """Class representing unsigned carry select adder.

    Carry select adder's logic is divided into a number of carry select blocks.
    Each block precomputes the outputs for both possible carry bit inputs for the
    given block of input indexes.

    Each carry select block is composed of these logic parts:
    Two full adder chains where the first accepts input Cin of 1 and the second accepts Cin of 0.
    (In case of first block only one chain is used and actual sum bits are obtained)
    From second block onward the blocks also contain multiplexers to choose the appropriate precomputed sum based
    on the carry out signal from the previous block (which serves as a select signal for the multiplexers).
    Finally the block also contains AND and OR gate to determine the carry out of the current block.

    ```
                           ┼   ┼                ┼   ┼
                   Cout┌───▼───▼───┐            │   │
                   ┌───┤    FAs    │◄─0         │   │
                   │   └─────┬─────┘            │   │
                   │         ┼ Sums             │   │
       ┌───────────┘         └─────────┐        │   │
       │                   ┼   ┼       │        │   │
    ┌──▼─┐  ┌─────┐Cout┌───▼───▼───┐   │Cout┌───▼───▼───┐
    │ OR │◄─┤ AND │◄───┤    FAs    │◄─1│  ┌─┤    FAs    │◄─Cin
    └─┬──┘  └───▲─┘    └─────┬─────┘   │  │ └─────┬─────┘
      │         │            ┼ Sums    │  │       ┼
      │         │          ┌─┘ ┌───────┘  │       │
      │         └──────────┼───┼───────┬──┘       │
      │                ┌───▼───▼───┐SEL│          │
      │                │ 2:1 MUXes │◄──┘          │
      │                └─────┬─────┘              │
      │                      ┼                    │
      ▼                      ▼                    ▼
     Cout                   Sums                 Sums
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        select_block_size (int, optional): Size of each composite select adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of unsigned csla. Defaults to "".
        name (str, optional): Name of unsigned csla. Defaults to "u_csla".
    """
    def __init__(self, a: Bus, b: Bus, select_block_size: int = 4, prefix: str = "", name: str = "u_csla", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        assert select_block_size > 0, "Block size of the individual select groups must be greater than 0."

        # To signify current number of blocks and number of bits that remain to be added into function blocks
        block_n = 0
        N_wires = self.N
        cin = ConstantWireValue0()
        while N_wires != 0:
            block_size = select_block_size if N_wires >= select_block_size else N_wires
            for i in range(block_size):
                if block_n == 0:  # First block contains just one chain of full adders (with Cin)
                    obj_adder = FullAdder(a=self.a.get_wire((block_n*select_block_size)+i), b=self.b.get_wire((block_n*select_block_size)+i), c=cin, prefix=self.prefix+f"_cin_block{block_n}_fa{i}")
                    self.add_component(obj_adder)
                    # Connecting adder's output sum bit to its proper position within the described circuit's output bus
                    self.out.connect(i, obj_adder.get_sum_wire())
                    # Updating cin for the the next adder/select block
                    cin = obj_adder.get_carry_wire()
                else:  # Other blocks contain two full adder chains where the first accepts input Cin of 1 and the second accepts Cin of 0; then 2-1 MUXes choose the proper output sum bit and finally the AND OR logic propagates carry signal into the next select block
                    if i == 0:
                        obj_adder0 = FullAdder(a=self.a.get_wire((block_n*select_block_size)+i), b=self.b.get_wire((block_n*select_block_size)+i), c=ConstantWireValue0(), prefix=self.prefix+f"_0block{block_n}_fa{i}")
                        obj_adder1 = FullAdder(a=self.a.get_wire((block_n*select_block_size)+i), b=self.b.get_wire((block_n*select_block_size)+i), c=ConstantWireValue1(), prefix=self.prefix+f"_1block{block_n}_fa{i}")
                    else:
                        obj_adder0 = FullAdder(a=self.a.get_wire((block_n*select_block_size)+i), b=self.b.get_wire((block_n*select_block_size)+i), c=self.get_previous_component(3).get_carry_wire(), prefix=self.prefix+f"_0block{block_n}_fa{i}")
                        obj_adder1 = FullAdder(a=self.a.get_wire((block_n*select_block_size)+i), b=self.b.get_wire((block_n*select_block_size)+i), c=self.get_previous_component(2).get_carry_wire(), prefix=self.prefix+f"_1block{block_n}_fa{i}")
                    self.add_component(obj_adder0)
                    self.add_component(obj_adder1)

                    self.add_component(TwoOneMultiplexer(a=obj_adder0.get_sum_wire(), b=obj_adder1.get_sum_wire(), c=cin, prefix=self.prefix+f"_block{block_n}_mux2to1_{i}"))
                    # Connecting the selected adder's output sum bit to its proper position within the described circuit's output bus
                    self.out.connect(i+(block_n*select_block_size), self.get_previous_component().out.get_wire())

                    if i == block_size-1:
                        # Updating cin for the the next select block
                        cout0 = obj_adder0.get_carry_wire()
                        cout1 = obj_adder1.get_carry_wire()
                        self.add_component(AndGate(a=cout1, b=cin, prefix=self.prefix+f"_block{block_n}_and{i}", parent_component=self))
                        self.add_component(OrGate(a=cout0, b=self.get_previous_component().out, prefix=self.prefix+f"_block{block_n}_or{i}", parent_component=self))
                        cin = self.get_previous_component().out

            N_wires -= block_size
            block_n += 1
        # Connection of final Cout
        self.out.connect(self.N, cin)


class SignedCarrySelectAdder(UnsignedCarrySelectAdder, GeneralCircuit):
    """Class representing signed carry select adder.

    Carry select adder's logic is divided into a number of carry select blocks.
    Each block precomputes the outputs for both possible carry bit inputs for the
    given block of input indexes.

    Each carry select block is composed of these logic parts:
    Two full adder chains where the first accepts input Cin of 1 and the second accepts Cin of 0.
    (In case of first block only one chain is used and actual sum bits are obtained)
    From second block onward the blocks also contain multiplexers to choose the appropriate precomputed sum based
    on the carry out signal from the previous block (which serves as a select signal for the multiplexers).
    Finally the block also contains AND and OR gate to determine the carry out of the current block.
    At last XOR gates are used to ensure proper sign extension.

    ```
                            ┼   ┼                ┼   ┼
                    Cout┌───▼───▼───┐            │   │
                    ┌───┤    FAs    │◄─0         │   │
                    │   └─────┬─────┘            │   │
                    │         ┼ Sums             │   │
        ┌───────────┘         └─────────┐        │   │
        │                   ┼   ┼       │        │   │
     ┌──▼─┐  ┌─────┐Cout┌───▼───▼───┐   │Cout┌───▼───▼───┐
     │ OR │◄─┤ AND │◄───┤    FAs    │◄─1│  ┌─┤    FAs    │◄─Cin
     └─┬──┘  └───▲─┘    └─────┬─────┘   │  │ └─────┬─────┘
       │         │            ┼ Sums    │  │       ┼
    ┌──▼───┐     │          ┌─┘ ┌───────┘  │       │
    │ SIGN │     └──────────┼───┼───────┬──┘       │
    │Extend│            ┌───▼───▼───┐SEL│          │
    └──┬───┘            │ 2:1 MUXes │◄──┘          │
       │                └─────┬─────┘              │
       │                      ┼                    │
       ▼                      ▼                    ▼
      Cout                   Sums                 Sums
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        select_block_size (int, optional): Size of each composite select adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of signed csla. Defaults to "".
        name (str, optional): Name of signed csla. Defaults to "s_csla".
    """
    def __init__(self, a: Bus, b: Bus, select_block_size: int = 4, prefix: str = "", name: str = "s_csla", **kwargs):
        # Signed bus length extension
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))
        
        super().__init__(a=a, b=b, select_block_size=select_block_size, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        self.add_component(XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self))
        self.add_component(XorGate(self.get_previous_component().out, self.out.get_wire(-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self))
        self.out.connect(self.N, self.get_previous_component().out)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    NandGate,
    NorGate,
    XorGate
)


class UnsignedArrayMultiplier(MultiplierCircuit):
    """Class representing unsigned array multiplier.

    Unsigned array multiplier represents N-bit multiplier composed of
    many AND gates and half/full adders to calculate partial products and
    gradually sum them.

    Downside is its rather big area because it is composed of many logic gates.

    ```
                                       A3B0     A2B0     A1B0     A0B0
                                       │ │      │ │      │ │      │ │
                                      ┌▼─▼┐    ┌▼─▼┐    ┌▼─▼┐    ┌▼─▼┐
                                      │AND│    │AND│    │AND│    │AND│
                                      └┬──┘    └┬──┘    └┬──┘    └─┬─┘
                                 A3B1  │  A2B1  │ A1B1   │  A0B1   │
                                ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐   │
                                │AND│  │ │AND│  │ │AND│  │ │AND│   │
                                └┬──┘  │ └┬──┘  │ └┬──┘  │ └┬──┘   │
                                 │     │  │     │  │     │  │      │
                             ┌───▼┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │
                             │    │   │    │   │    │   │    │     │
                     ┌───────┤ HA │◄──┤ FA │◄──┤ FA │◄──┤ HA │     │
                     │       │    │   │    │   │    │   │    │     │
                     │       └┬───┘   └┬───┘   └┬───┘   └─┬──┘     │
                     │  A3B2  │  A2B2  │  A1B2  │  A0B2   │        │
                     │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐   │        │
                     │ │AND│  │ │AND│  │ │AND│  │ │AND│   │        │
                     │ └┬──┘  │ └┬──┘  │ └┬──┘  │ └┬──┘   │        │
                     │  │     │  │     │  │     │  │      │        │
                    ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │        │
                    │    │   │    │   │    │   │    │     │        │
            ┌───────┤ FA │◄──┤ FA │◄──┤ FA │◄──┤ HA │     │        │
            │       │    │   │    │   │    │   │    │     │        │
            │       └┬───┘   └┬───┘   └┬───┘   └─┬──┘     │        │
            │  A3B3  │  A2B3  │  A1B3  │  A0B3   │        │        │
            │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐   │        │        │
            │ │AND│  │ │AND│  │ │AND│  │ │AND│   │        │        │
            │ └┬──┘  │ └┬──┘  │ └┬──┘  │ └┬──┘   │        │        │
            │  │     │  │     │  │     │  │      │        │        │
           ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │        │        │
           │    │   │    │   │    │   │    │     │        │        │
    ┌──────┤ FA │◄──┤ FA │◄──┤ FA │◄──┤ HA │     │        │        │
    │      │    │   │    │   │    │   │    │     │        │        │
    │      └─┬──┘   └─┬──┘   └─┬──┘   └─┬──┘     │        │        │
    │        │        │        │        │        │        │        │
    ▼        ▼        ▼        ▼        ▼        ▼        ▼        ▼
    P7       P6       P5       P4       P3       P2       P1       P0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned array multiplier. Defaults to "".
        name (str, optional): Name of unsigned array multiplier. Defaults to "u_arrmul".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_arrmul", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Gradual generation of partial products
        for b_multiplier_index in range(self.N):
            for a_multiplicand_index in range(self.N):
                # AND gates generation for calculation of partial products
                obj_and = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                self.add_component(obj_and)

                if b_multiplier_index != 0:
                    previous_product = self.components[a_multiplicand_index + b_multiplier_index].out if b_multiplier_index == 1 else self.get_previous_partial_product(a_index=a_multiplicand_index, b_index=b_multiplier_index)
                    # HA generation for first 1-bit adder in each row starting from the second one
                    if a_multiplicand_index == 0:
                        obj_adder = HalfAdder(self.get_previous_component().out, previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)
                        # Product generation
                        self.out.connect(b_multiplier_index, obj_adder.get_sum_wire())

                    # HA generation, last 1-bit adder in second row
                    elif a_multiplicand_index == self.N-1 and b_multiplier_index == 1:
                        obj_adder = HalfAdder(self.get_previous_component().out, self.get_previous_component(number=2).get_carry_wire(), prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)

                    # FA generation
                    else:
                        obj_adder = FullAdder(self.get_previous_component().out, previous_product, self.get_previous_component(number=2).get_carry_wire(), prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)

                # PRODUCT GENERATION
                if a_multiplicand_index == 0 and b_multiplier_index == 0:
                    self.out.connect(a_multiplicand_index, obj_and.out)

                    # 1 bit multiplier case
                    if a_multiplicand_index == self.N-1:
                        self.out.connect(a_multiplicand_index+1, ConstantWireValue0())

                elif b_multiplier_index == self.N-1:
                    self.out.connect(b_multiplier_index + a_multiplicand_index, obj_adder.get_sum_wire())

                    if a_multiplicand_index == self.N-1:
                        self.out.connect(self.out.N-1, obj_adder.get_carry_wire())


class SignedArrayMultiplier(MultiplierCircuit):
    """Class representing signed array multiplier.

    Signed array multiplier represents N-bit multiplier composed of
    many AND/NAND gates and half/full adders to calculate partial products and
    gradually sum them.

    Downside is its rather big area because it is composed of many logic gates.

    ```
                                         A3B0     A2B0     A1B0     A0B0
                                         │ │      │ │      │ │      │ │
                                        ┌▼─▼─┐   ┌▼─▼┐    ┌▼─▼┐    ┌▼─▼┐
                                        │NAND│   │AND│    │AND│    │AND│
                                        └┬───┘   └┬──┘    └┬──┘    └─┬─┘
                                   A3B1  │  A2B1  │ A1B1   │  A0B1   │
                                  ┌▼─▼─┐ │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐   │
                                  │NAND│ │ │AND│  │ │AND│  │ │AND│   │
                                1 └┬───┘ │ └┬──┘  │ └┬──┘  │ └┬──┘   │
                                │  │     │  │     │  │     │  │      │
                               ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │
                               │    │   │    │   │    │   │    │     │
                       ┌───────┤ FA │◄──┤ FA │◄──┤ FA │◄──┤ HA │     │
                       │       │    │   │    │   │    │   │    │     │
                       │       └┬───┘   └┬───┘   └┬───┘   └─┬──┘     │
                       │  A3B2  │  A2B2  │  A1B2  │  A0B2   │        │
                       │ ┌▼─▼─┐ │ ┌▼─▼┐  │ ┌▼─▼┐  │ ┌▼─▼┐   │        │
                       │ │NAND│ │ │AND│  │ │AND│  │ │AND│   │        │
                       │ └┬───┘ │ └┬──┘  │ └┬──┘  │ └┬──┘   │        │
                       │  │     │  │     │  │     │  │      │        │
                      ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │        │
                      │    │   │    │   │    │   │    │     │        │
              ┌───────┤ FA │◄──┤ FA │◄──┤ FA │◄──┤ HA │     │        │
              │       │    │   │    │   │    │   │    │     │        │
              │       └┬───┘   └┬───┘   └┬───┘   └─┬──┘     │        │
              │  A3B3  │  A2B3  │  A1B3  │  A0B3   │        │        │
              │ ┌▼─▼┐  │ ┌▼─▼─┐ │ ┌▼─▼─┐ │ ┌▼─▼─┐  │        │        │
              │ │AND│  │ │NAND│ │ │NAND│ │ │NAND│  │        │        │
      1       │ └┬──┘  │ └┬───┘ │ └┬───┘ │ └┬───┘  │        │        │
      │       │  │     │  │     │  │     │  │      │        │        │
    ┌─▼──┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐   ┌▼──▼┐     │        │        │
    │    │   │    │   │    │   │    │   │    │     │        │        │
    │XOR │◄──┤ FA │◄──┤ FA │◄──┤ FA │◄──┤ HA │     │        │        │
    │    │   │    │   │    │   │    │   │    │     │        │        │
    └─┬──┘   └─┬──┘   └─┬──┘   └─┬──┘   └─┬──┘     │        │        │
      │        │        │        │        │        │        │        │
      ▼        ▼        ▼        ▼        ▼        ▼        ▼        ▼
      P7       P6       P5       P4       P3       P2       P1       P0
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed array multiplier. Defaults to "".
        name (str, optional): Name of signed array multiplier. Defaults to "s_arrmul".
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_arrmul", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, signed=True, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix, desired_extension_wire=self.a.get_wire(self.a.N-1))
        self.b.bus_extend(N=self.N, prefix=b.prefix, desired_extension_wire=self.b.get_wire(self.b.N-1))

        # Gradual generation of partial products
        for b_multiplier_index in range(self.N):
            for a_multiplicand_index in range(self.N):
                # AND and NAND gates generation for calculation of partial products and sign extension
                if (b_multiplier_index == self.N-1 and a_multiplicand_index != self.N-1) or (b_multiplier_index != self.N-1 and a_multiplicand_index == self.N-1):
                    obj_nand = NandGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_nand"+str(a_multiplicand_index)+"_"+str(b_multiplier_index), parent_component=self)
                    self.add_component(obj_nand)
                else:
                    obj_and = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index), parent_component=self)
                    self.add_component(obj_and)

                if b_multiplier_index != 0:
                    previous_product = self.components[a_multiplicand_index + b_multiplier_index].out if b_multiplier_index == 1 else self.get_previous_partial_product(a_index=a_multiplicand_index, b_index=b_multiplier_index)
                    # HA generation for first 1-bit adder in each row starting from the second one
                    if a_multiplicand_index == 0:
                        obj_adder = HalfAdder(self.get_previous_component().out, previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)
                        # Product generation
                        self.out.connect(b_multiplier_index, obj_adder.get_sum_wire())

                    # FA generation
                    else:
                        # Constant wire with value 1 used at the last FA in second row (as one of its inputs) for signed multiplication (based on Baugh Wooley algorithm)
                        if a_multiplicand_index == self.N-1 and b_multiplier_index == 1:
                            previous_product = ConstantWireValue1()

                        obj_adder = FullAdder(self.get_previous_component().out, previous_product, self.get_previous_component(number=2).get_carry_wire(), prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)

                # PRODUCT GENERATION
                if a_multiplicand_index == 0 and b_multiplier_index == 0:
                    self.out.connect(a_multiplicand_index, obj_and.out)

                    # 1 bit multiplier case
                    if a_multiplicand_index == self.N-1:
                        obj_nor = NorGate(ConstantWireValue1(), self.get_previous_component().out, prefix=self.prefix+"_nor_zero_extend", parent_component=self)
                        self.add_component(obj_nor)

                        self.out.connect(a_multiplicand_index+1, obj_nor.out)

                elif b_multiplier_index == self.N-1:
                    self.out.connect(b_multiplier_index + a_multiplicand_index, obj_adder.get_sum_wire())

                    if a_multiplicand_index == self.N-1:
                        obj_xor = XorGate(self.get_previous_component().get_carry_wire(), ConstantWireValue1(), prefix=self.prefix+"_xor"+str(a_multiplicand_index+1)+"_"+str(b_multiplier_index), parent_component=self)
                        self.add_component(obj_xor)

                        self.out.connect(self.out.N-1, obj_xor.out)



import os
import sys
# Add the parent directory to the system path
DIR_PATH = os.path.dirname(os.path.abspath(__file__))
sys.path.insert(0, os.path.join(DIR_PATH, '..'))

import numpy as np
import math
from io import StringIO

from ariths_gen.wire_components import (
    Wire,
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)

from ariths_gen.core.arithmetic_circuits import GeneralCircuit

from ariths_gen.multi_bit_circuits.adders import (
    UnsignedCarryLookaheadAdder,
    UnsignedPGRippleCarryAdder,
    UnsignedRippleCarryAdder,
    SignedCarryLookaheadAdder,
    SignedPGRippleCarryAdder,
    SignedRippleCarryAdder,
    UnsignedCarrySkipAdder,
    SignedCarrySkipAdder,
    UnsignedKoggeStoneAdder,
    SignedKoggeStoneAdder,
    UnsignedBrentKungAdder,
    SignedBrentKungAdder,
    UnsignedSklanskyAdder,
    SignedSklanskyAdder,
    UnsignedHanCarlsonAdder,
    SignedHanCarlsonAdder,
    UnsignedLadnerFischerAdder,
    SignedLadnerFischerAdder,
    UnsignedKnowlesAdder,
    SignedKnowlesAdder,
    UnsignedCarrySelectAdder,
    SignedCarrySelectAdder,
    UnsignedConditionalSumAdder,
    SignedConditionalSumAdder,
    UnsignedCarryIncrementAdder,
    SignedCarryIncrementAdder
)

from ariths_gen.multi_bit_circuits.subtractors import (
    UnsignedRippleBorrowSubtractor,
    UnsignedRippleCarrySubtractor,
    SignedRippleBorrowSubtractor,
    SignedRippleCarrySubtractor
)

from ariths_gen.multi_bit_circuits.multipliers import (
    UnsignedDaddaMultiplier,
    UnsignedArrayMultiplier,
    UnsignedWallaceMultiplier,
    UnsignedCarrySaveMultiplier,
    SignedArrayMultiplier,
    SignedDaddaMultiplier,
    SignedWallaceMultiplier,
    SignedCarrySaveMultiplier
)


from ariths_gen.multi_bit_circuits.approximate_multipliers import (
    UnsignedTruncatedArrayMultiplier,
    UnsignedTruncatedCarrySaveMultiplier,
    UnsignedBrokenArrayMultiplier,
    UnsignedBrokenCarrySaveMultiplier
)

from ariths_gen.core.cgp_circuit import UnsignedCGPCircuit, SignedCGPCircuit


def test_cgp_unsigned_add():
    """ Test unsigned adders """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N)
    bv = av.reshape(-1, 1)
    expected = av + bv

    # Non configurable multi-bit adders
    for c in [UnsignedPGRippleCarryAdder, UnsignedRippleCarryAdder, UnsignedConditionalSumAdder, UnsignedKoggeStoneAdder, UnsignedBrentKungAdder, UnsignedSklanskyAdder]:
        add = c(a, b)
        code = StringIO()
        add.get_cgp_code_flat(code)
        cgp_code = code.getvalue()
        print(cgp_code)

        add2 = UnsignedCGPCircuit(cgp_code, [N, N])
        o = StringIO()
        add2.get_v_code_flat(o)
        print(o.getvalue())

        r = add2(av, bv)
        assert add(0, 0) == 0
        assert add2(0, 0) == 0
        np.testing.assert_array_equal(expected, r)

    # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the adder will use the block_size argument it recognizes, others are ignored)
    for c in [UnsignedCarryLookaheadAdder, UnsignedCarrySkipAdder, UnsignedCarrySelectAdder, UnsignedCarryIncrementAdder]:
        for bs in range(1, N+1):
            add = c(a, b, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
            r = add(av, bv)
            code = StringIO()
            add.get_cgp_code_flat(code)
            cgp_code = code.getvalue()
            print(cgp_code)

            add2 = UnsignedCGPCircuit(cgp_code, [N, N])
            o = StringIO()
            add2.get_v_code_flat(o)
            print(o.getvalue())

            r = add2(av, bv)
            assert add(0, 0) == 0
            assert add2(0, 0) == 0
            np.testing.assert_array_equal(expected, r)

    # Multi-bit tree adders with configurable structure based on input bit width (2 configs tested here for the 9-bitwidth input)
    for c in [UnsignedHanCarlsonAdder, UnsignedKnowlesAdder, UnsignedLadnerFischerAdder]:
        for config in range(1, (math.ceil(math.log(N, 2))-2)+1):
            add = c(a, b, config_choice=config)
            r = add(av, bv)
            code = StringIO()
            add.get_cgp_code_flat(code)
            cgp_code = code.getvalue()
            print(cgp_code)

            add2 = UnsignedCGPCircuit(cgp_code, [N, N])
            o = StringIO()
            add2.get_v_code_flat(o)
            print(o.getvalue())

            r = add2(av, bv)
            assert add(0, 0) == 0
            assert add2(0, 0) == 0
            np.testing.assert_array_equal(expected, r)


def test_cgp_signed_add():
    """ Test signed adders """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(-(2**(N-1)), 2**(N-1))
    bv = av.reshape(-1, 1)
    expected = av + bv

    # Non configurable multi-bit adders
    for c in [SignedPGRippleCarryAdder, SignedRippleCarryAdder, SignedConditionalSumAdder, SignedKoggeStoneAdder, SignedBrentKungAdder, SignedSklanskyAdder]:
        add = c(a, b)
        r = add(av, bv)
        code = StringIO()
        add.get_cgp_code_flat(code)
        cgp_code = code.getvalue()
        print(cgp_code)

        add2 = SignedCGPCircuit(cgp_code, [N, N])
        o = StringIO()
        add2.get_v_code_flat(o)
        print(o.getvalue())

        r = add2(av, bv)
        assert add(0, 0) == 0
        assert add2(0, 0) == 0
        np.testing.assert_array_equal(expected, r)

    # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the adder will use the block_size argument it recognizes, others are ignored)
    for c in [SignedCarryLookaheadAdder, SignedCarrySkipAdder, SignedCarrySelectAdder, SignedCarryIncrementAdder]:
        for bs in range(1, N+1):
            add = c(a, b, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
            r = add(av, bv)
            code = StringIO()
            add.get_cgp_code_flat(code)
            cgp_code = code.getvalue()
            print(cgp_code)

            add2 = SignedCGPCircuit(cgp_code, [N, N])
            o = StringIO()
            add2.get_v_code_flat(o)
            print(o.getvalue())

            r = add2(av, bv)
            assert add(0, 0) == 0
            assert add2(0, 0) == 0
            np.testing.assert_array_equal(expected, r)

    # Multi-bit tree adders with configurable structure based on input bit width (2 configs tested here for the 9-bitwidth input)
    for c in [SignedHanCarlsonAdder, SignedKnowlesAdder, SignedLadnerFischerAdder]:
        for config in range(1, (math.ceil(math.log(N, 2))-2)+1):
            add = c(a, b, config_choice=config)
            r = add(av, bv)
            code = StringIO()
            add.get_cgp_code_flat(code)
            cgp_code = code.getvalue()
            print(cgp_code)

            add2 = SignedCGPCircuit(cgp_code, [N, N])
            o = StringIO()
            add2.get_v_code_flat(o)
            print(o.getvalue())

            r = add2(av, bv)
            assert add(0, 0) == 0
            assert add2(0, 0) == 0
            np.testing.assert_array_equal(expected, r)


def test_cgp_unsigned_sub():
    """ Test unsigned subtractors """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N)
    bv = av.reshape(-1, 1)
    expected = av - bv

    #for c in [UnsignedRippleBorrowSubtractor, UnsignedRippleCarrySubtractor]:
    for c in [UnsignedRippleBorrowSubtractor]:
        sub = c(a, b)
        code = StringIO()
        sub.get_cgp_code_flat(code)
        cgp_code = code.getvalue()

        sub2 = UnsignedCGPCircuit(cgp_code, [N, N], signed_out=True)
        o = StringIO()
        sub2.get_v_code_flat(o)
        print(o.getvalue())

        r = sub2(av, bv)
        assert sub(0, 0) == 0
        assert sub2(0, 0) == 0
        np.testing.assert_array_equal(expected, r)
    

def test_cgp_signed_sub():
    """ Test signed subtractors """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(-(2**(N-1)), 2**(N-1))
    bv = av.reshape(-1, 1)
    expected = av - bv

    for c in [SignedRippleBorrowSubtractor, SignedRippleCarrySubtractor]:
        sub = c(a, b)
        r = sub(av, bv)
        code = StringIO()
        sub.get_cgp_code_flat(code)
        cgp_code = code.getvalue()
        print(cgp_code)

        sub2 = SignedCGPCircuit(cgp_code, [N, N])
        o = StringIO()
        sub2.get_v_code_flat(o)
        print(o.getvalue())

        r = sub2(av, bv)
        assert sub(0, 0) == 0
        assert sub2(0, 0) == 0
        np.testing.assert_array_equal(expected, r)


def test_cgp_unsigned_mul():
    """ Test unsigned multipliers """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(2**N)
    bv = av.reshape(-1, 1)
    expected = av * bv

    # No configurability
    for c in [UnsignedArrayMultiplier]:
        mul = c(a, b)
        code = StringIO()
        mul.get_cgp_code_flat(code)
        cgp_code = code.getvalue()

        mul2 = UnsignedCGPCircuit(cgp_code, [N, N])
        r = mul2(av, bv)

        assert mul(0, 0) == 0
        assert mul2(0, 0) == 0
        np.testing.assert_array_equal(expected, r)

    # Configurable PPA
    for c in [UnsignedDaddaMultiplier, UnsignedCarrySaveMultiplier, UnsignedWallaceMultiplier]:
        # Non configurable multi-bit adders
        for ppa in [UnsignedPGRippleCarryAdder, UnsignedRippleCarryAdder, UnsignedConditionalSumAdder, UnsignedKoggeStoneAdder, UnsignedBrentKungAdder, UnsignedSklanskyAdder]:
            # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
            if c == UnsignedWallaceMultiplier:
                mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False)
                code = StringIO()
                mul.get_cgp_code_flat(code)
                cgp_code = code.getvalue()

                mul2 = UnsignedCGPCircuit(cgp_code, [N, N])
                r = mul2(av, bv)

                assert mul(0, 0) == 0
                assert mul2(0, 0) == 0
                np.testing.assert_array_equal(expected, r)

            mul = c(a, b, unsigned_adder_class_name=ppa)
            code = StringIO()
            mul.get_cgp_code_flat(code)
            cgp_code = code.getvalue()

            mul2 = UnsignedCGPCircuit(cgp_code, [N, N])
            r = mul2(av, bv)

            assert mul(0, 0) == 0
            assert mul2(0, 0) == 0
            np.testing.assert_array_equal(expected, r)

        # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the ppa will use the block_size argument it recognizes, others are ignored)
        for ppa in [UnsignedCarryLookaheadAdder, UnsignedCarrySkipAdder, UnsignedCarrySelectAdder, UnsignedCarryIncrementAdder]:
            for i in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, cla_block_size=i, bypass_block_size=i, select_block_size=i, increment_block_size=i)
                    code = StringIO()
                    mul.get_cgp_code_flat(code)
                    cgp_code = code.getvalue()

                    mul2 = UnsignedCGPCircuit(cgp_code, [N, N])
                    r = mul2(av, bv)

                    assert mul(0, 0) == 0
                    assert mul2(0, 0) == 0
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, cla_block_size=i, bypass_block_size=i, select_block_size=i, increment_block_size=i)
                code = StringIO()
                mul.get_cgp_code_flat(code)
                cgp_code = code.getvalue()

                mul2 = UnsignedCGPCircuit(cgp_code, [N, N])
                r = mul2(av, bv)

                assert mul(0, 0) == 0
                assert mul2(0, 0) == 0
                np.testing.assert_array_equal(expected, r)

        # Multi-bit tree adders with configurable structure based on input bit width (NOTE for showcase here, the second config would be applicable from bit width 9 onward; not tested here for the sake of saving deployment testing time)
        for adder in [UnsignedHanCarlsonAdder, UnsignedKnowlesAdder, UnsignedLadnerFischerAdder]:
            for i in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, config_choice=1)
                    code = StringIO()
                    mul.get_cgp_code_flat(code)
                    cgp_code = code.getvalue()

                    mul2 = UnsignedCGPCircuit(cgp_code, [N, N])
                    r = mul2(av, bv)

                    assert mul(0, 0) == 0
                    assert mul2(0, 0) == 0
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, config_choice=1)
                code = StringIO()
                mul.get_cgp_code_flat(code)
                cgp_code = code.getvalue()

                mul2 = UnsignedCGPCircuit(cgp_code, [N, N])
                r = mul2(av, bv)

                assert mul(0, 0) == 0
                assert mul2(0, 0) == 0
                np.testing.assert_array_equal(expected, r)


def test_cgp_signed_mul():
    """ Test signed multipliers """
    N = 7
    a = Bus(N=N, prefix="a")
    b = Bus(N=N, prefix="b")
    av = np.arange(-(2**(N-1)), 2**(N-1))
    bv = av.reshape(-1, 1)
    expected = av * bv

    # No configurability
    for c in [SignedArrayMultiplier]:
        mul = c(a, b)
        code = StringIO()
        mul.get_cgp_code_flat(code)
        cgp_code = code.getvalue()

        mul2 = SignedCGPCircuit(cgp_code, [N, N])
        r = mul2(av, bv)

        assert mul(0, 0) == 0
        assert mul2(0, 0) == 0
        np.testing.assert_array_equal(expected, r)

    # Configurable PPA
    for c in [SignedDaddaMultiplier, SignedCarrySaveMultiplier, SignedWallaceMultiplier]:
        # Non configurable multi-bit adders
        for ppa in [UnsignedPGRippleCarryAdder, UnsignedRippleCarryAdder, UnsignedConditionalSumAdder, UnsignedKoggeStoneAdder, UnsignedBrentKungAdder, UnsignedSklanskyAdder]:
            # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
            if c == UnsignedWallaceMultiplier:
                mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False)
                code = StringIO()
                mul.get_cgp_code_flat(code)
                cgp_code = code.getvalue()

                mul2 = SignedCGPCircuit(cgp_code, [N, N])
                r = mul2(av, bv)

                assert mul(0, 0) == 0
                assert mul2(0, 0) == 0
                np.testing.assert_array_equal(expected, r)

            mul = c(a, b, unsigned_adder_class_name=ppa)
            code = StringIO()
            mul.get_cgp_code_flat(code)
            cgp_code = code.getvalue()

            mul2 = SignedCGPCircuit(cgp_code, [N, N])
            r = mul2(av, bv)

            assert mul(0, 0) == 0
            assert mul2(0, 0) == 0
            np.testing.assert_array_equal(expected, r)

        # Multi-bit adders with configurable (uniform) logic blocks for parallel prefix computation (the ppa will use the block_size argument it recognizes, others are ignored)
        for ppa in [UnsignedCarryLookaheadAdder, UnsignedCarrySkipAdder, UnsignedCarrySelectAdder, UnsignedCarryIncrementAdder]:
            for bs in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
                    code = StringIO()
                    mul.get_cgp_code_flat(code)
                    cgp_code = code.getvalue()

                    mul2 = SignedCGPCircuit(cgp_code, [N, N])
                    r = mul2(av, bv)

                    assert mul(0, 0) == 0
                    assert mul2(0, 0) == 0
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, cla_block_size=bs, bypass_block_size=bs, select_block_size=bs, increment_block_size=bs)
                code = StringIO()
                mul.get_cgp_code_flat(code)
                cgp_code = code.getvalue()

                mul2 = SignedCGPCircuit(cgp_code, [N, N])
                r = mul2(av, bv)

                assert mul(0, 0) == 0
                assert mul2(0, 0) == 0
                np.testing.assert_array_equal(expected, r)

        # Multi-bit tree adders with configurable structure based on input bit width (NOTE for showcase here, the second config would be applicable from bit width 9 onward; not tested here for the sake of saving deployment testing time)
        for ppa in [UnsignedHanCarlsonAdder, UnsignedKnowlesAdder, UnsignedLadnerFischerAdder]:
            for i in range(1, N+1):
                # Test first the array wallace tree implementation (using more HAs/FAs than CSA implementation)
                if c == UnsignedWallaceMultiplier:
                    mul = c(a, b, unsigned_adder_class_name=ppa, use_csa=False, config_choice=1)
                    code = StringIO()
                    mul.get_cgp_code_flat(code)
                    cgp_code = code.getvalue()

                    mul2 = SignedCGPCircuit(cgp_code, [N, N])
                    r = mul2(av, bv)

                    assert mul(0, 0) == 0
                    assert mul2(0, 0) == 0
                    np.testing.assert_array_equal(expected, r)

                mul = c(a, b, unsigned_adder_class_name=ppa, config_choice=1)
                code = StringIO()
                mul.get_cgp_code_flat(code)
                cgp_code = code.getvalue()

                mul2 = SignedCGPCircuit(cgp_code, [N, N])
                r = mul2(av, bv)

                assert mul(0, 0) == 0
                assert mul2(0, 0) == 0
                np.testing.assert_array_equal(expected, r)


def test_cgp_variant1():
    # one input is connected to the output (first bit)
    cgp = "{16,9,37,1,2,1,0}([18]15,12,1)([19]7,7,4)([20]3,12,5)([21]17,3,0)([22]8,14,3)([23]15,3,6)([24]14,0,2)([25]9,9,5)([26]17,13,1)([27]12,13,0)([28]7,16,8)([29]12,11,0)([30]5,13,3)([31]5,13,2)([32]30,12,5)([33]30,29,2)([34]31,33,3)([35]6,14,4)([36]6,14,2)([37]35,34,4)([38]35,34,2)([39]36,38,3)([40]7,15,4)([41]7,15,2)([42]40,39,4)([43]40,39,2)([44]41,43,3)([45]8,16,4)([46]8,16,2)([47]45,44,4)([48]45,44,2)([49]46,48,3)([50]9,17,4)([51]9,17,2)([52]50,49,4)([53]50,49,2)([54]51,53,3)(11,40,33,32,37,42,47,52,54)"

    c = UnsignedCGPCircuit(cgp, [8, 8], name="cgp_circuit")
    assert c(0, 0) == 8  # TypeError: 'int' object is not subscriptable


if __name__ == "__main__":
    test_cgp_unsigned_add()
    test_cgp_signed_add()
    test_cgp_unsigned_sub()
    test_cgp_signed_sub()
    test_cgp_unsigned_mul()
    test_cgp_signed_mul()
    test_cgp_variant1()
    print("CGP Python tests were successful!")



from ariths_gen.core.one_bit_circuits import TwoInputOneBitCircuit
from ariths_gen.one_bit_circuits.logic_gates import AndGate, NandGate, OrGate, NorGate, XorGate, XnorGate, NotGate
from ariths_gen.wire_components import Wire, Bus


class HalfAdder(TwoInputOneBitCircuit):
    """Class representing two input one bit half adder.

    ```
        ┌──────┐
    ───►│      ├─► Sum
        │      │
    ───►│      ├─► Cout
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        prefix (str, optional): Prefix name of half adder. Defaults to "".
        name (str, optional): Name of half adder. Defaults to "ha".
    """
    use_verilog_instance = False

    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), prefix: str = "", name: str = "ha"):
        super().__init__(a, b, prefix=prefix, name=name)
        # 2 wires for component's bus output (sum, cout)
        self.out = Bus(self.prefix+"_out", 2)

        # Sum
        # XOR gate for calculation of 1-bit sum
        obj_xor = XorGate(a, b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=0, parent_component=self)
        self.add_component(obj_xor)
        self.out.connect(0, obj_xor.out)

        # Cout
        # AND gate for calculation of 1-bit cout
        obj_and = AndGate(a, b, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), outid=1, parent_component=self)
        self.add_component(obj_and)
        self.out.connect(1, obj_and.out)

    def get_sum_wire(self):
        """Get output wire carrying sum value.

        Returns:
           Wire: Return sum wire.
        """
        return self.out.get_wire(0)

    def get_carry_wire(self):
        """Get output wire carrying carry out value.

        Returns:
           Wire: Return carry out wire.
        """
        return self.out.get_wire(1)

    def get_init_v_flat(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_init_v_flat()

        return "  " + self.use_verilog_instance.format(
            **{
                "unit": self.prefix,
                "wirea": f"1'b{self.a.value}" if self.a.is_const() else self.a.name,
                "wireb": f"1'b{self.b.value}" if self.b.is_const() else self.b.name,
                "wireys": self.get_sum_wire().prefix,
                "wireyc": self.get_carry_wire().prefix,
            }) + ";\n"

    def get_self_init_v_hier(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_self_init_v_hier()

        unique_out_wires = []
        for o in self.out.bus:
            unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name)

        return "  " + self.use_verilog_instance.format(
            **{
                "unit": self.prefix,
                "wirea": self.a.name,
                "wireb": self.b.name,
                "wireys": unique_out_wires[0],
                "wireyc": unique_out_wires[1],
            }) + ";\n"

    def get_circuit_v(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_circuit_v()

        return f"{self.get_prototype_v_hier()}" + \
               f"{self.get_self_init_v_hier()}" + \
               f"endmodule"


class PGLogicBlock(TwoInputOneBitCircuit):
    """Class representing two input one bit propagate/generate logic block.

    ```
        ┌──────┐
    ───►│      ├─► P
        │      ├─► G
    ───►│      ├─► S
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        prefix (str, optional): Prefix name of pg logic block. Defaults to "".
        name (str, optional): Name of pg logic block. Defaults to "pg_logic".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), prefix: str = "", name: str = "pg_logic"):
        super().__init__(a, b, prefix=prefix, name=name)
        # 3 wires for component's bus output (propagate, generate, sum)
        self.out = Bus(self.prefix+"_out", 3)

        # PG logic
        propagate_or = OrGate(a, b, prefix=self.prefix+"_or"+str(self.get_instance_num(cls=OrGate)), outid=0, parent_component=self)
        self.add_component(propagate_or)
        generate_and = AndGate(a, b, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), outid=1, parent_component=self)
        self.add_component(generate_and)
        sum_xor = XorGate(a, b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=2, parent_component=self)
        self.add_component(sum_xor)

        self.out.connect(0, propagate_or.out)
        self.out.connect(1, generate_and.out)
        self.out.connect(2, sum_xor.out)

    def get_propagate_wire(self):
        """Get output wire carrying propagate signal value.

        Returns:
           Wire: Return propagate wire.
        """
        return self.out.get_wire(0)

    def get_generate_wire(self):
        """Get output wire carrying generate signal value.

        Returns:
           Wire: Return generate wire.
        """
        return self.out.get_wire(1)

    def get_sum_wire(self):
        """Get output wire carrying sum value.

        Returns:
           Wire: Return sum wire.
        """
        return self.out.get_wire(2)


class HalfSubtractor(TwoInputOneBitCircuit):
    """Class representing two input one bit half subtractor.

    ```
        ┌──────┐
    ───►│      ├─► Difference
        │      │
    ───►│      ├─► Bout
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        prefix (str, optional): Prefix name of half subtractor adder. Defaults to "".
        name (str, optional): Name of half subtractor adder. Defaults to "hs".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), prefix: str = "", name: str = "hs"):
        super().__init__(a, b, prefix=prefix, name=name)
        # 2 wires for component's bus output (difference, bout)
        self.out = Bus(self.prefix+"_out", 2)

        # Difference
        # XOR gate for calculation of 1-bit difference
        difference_xor = XorGate(a=self.a, b=self.b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=0, parent_component=self)
        self.add_component(difference_xor)
        self.out.connect(0, difference_xor.out)

        # Bout
        # NOT and AND gates for calculation of 1-bit borrow out
        not_obj = NotGate(a=self.a, prefix=self.prefix+"_not"+str(self.get_instance_num(cls=NotGate)), parent_component=self)
        self.add_component(not_obj)

        borrow_and = AndGate(a=not_obj.out, b=self.b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=1, parent_component=self)
        self.add_component(borrow_and)
        self.out.connect(1, borrow_and.out)

    def get_difference_wire(self):
        """Get output wire carrying difference value.

        Returns:
           Wire: Return difference wire.
        """
        return self.out.get_wire(0)

    def get_borrow_wire(self):
        """Get output wire carrying borrow out value.

        Returns:
           Wire: Return borrow out wire.
        """
        return self.out.get_wire(1)



from .general_circuit import (
    GeneralCircuit
)
from ariths_gen.wire_components import (
    Bus
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    NandGate
)
import math


class MultiplierCircuit(GeneralCircuit):
    """Class represents a general multiplier circuit derived from `GeneralCircuit` class.

    The __init__ method calls parent class __init__ method which fills some mandatory attributes concerning general circuit
    that are later used for generation into various representations.
    """
    def __init__(self, prefix: str, name: str, out_N: int, inner_component: bool = False, inputs: list = [], one_bit_circuit: bool = False, signed: bool = False, outname: str = "", **kwargs):
        super().__init__(prefix=prefix, name=name, out_N=out_N, inner_component=inner_component, inputs=inputs, one_bit_circuit=one_bit_circuit, signed=signed, outname=outname, **kwargs)

    # Array/approximate multipliers
    def get_previous_partial_product(self, a_index: int, b_index: int, mult_type=""):
        """Used in array and approximate multipliers to get previous row's component output wires for further connection to another component's input.

        Args:
            a_index (int): First input wire index.
            b_index (int): Second input wire index.
            mult_type (string, optional): Specifies what type of multiplier circuit has called this method. It is used for proper retrieval of index into the components list to allow appropriate interconnection of the multiplier circuit's inner subcomponents. It expects "" for ordinary multipliers, `bam` or `tm` for specific approximate multipliers. Defaults to "".

        Returns:
            Wire: Previous row's component wire of corresponding pp.
        """
        # To get the index of previous row's connecting adder and its generated pp
        if mult_type == "bam":
            # TODO alter to be more compact
            ids_sum = 0
            for row in range(self.horizontal_cut + self.ommited_rows, b_index):
                first_row_elem_id = self.vertical_cut-row if self.vertical_cut-row > 0 else 0
                # First pp row composed just from gates
                if row == self.horizontal_cut + self.ommited_rows:
                    # Minus one because the first component has index 0 instead of 1
                    ids_sum += sum([1 for gate_pos in range(first_row_elem_id, self.N)])-1
                elif row == b_index-1:
                    ids_sum += sum([2 for gate_adder_pos in range(first_row_elem_id, self.N) if gate_adder_pos <= a_index+1])
                else:
                    ids_sum += sum([2 for gate_adder_pos in range(first_row_elem_id, self.N)])
            # Index calculation should be redone, but it works even this way
            index = ids_sum+2 if a_index == self.N-1 else ids_sum
        elif mult_type == "tm":
            index = ((b_index-self.truncation_cut-2) * ((self.N-self.truncation_cut)*2)) + ((self.N-self.truncation_cut-1)+2*(a_index-self.truncation_cut+2))
        else:
            index = ((b_index-2) * ((self.N)*2)) + ((self.N-1)+2*(a_index+2))

        # Get carry wire as input for the last adder in current row
        if a_index == self.N-1:
            index = index-2
            return self.components[index].get_carry_wire()
        # Get sum wire as input for current adder
        else:
            return self.components[index].get_sum_wire()

    # Dadda/Wallace multipliers
    @staticmethod
    def get_maximum_height(initial_value: int):
        """Used in dadda multipliers to get multiplier's maximum height.

        Maximum height sequence as defined here: https://en.wikipedia.org/wiki/Dadda_multiplier
        d(j=1) = 2; d(j+1) = floor(1.5*d)

        `j` stands for initial stage value
        `d` stands for maximum height for current initial stage value

        Args:
            initial_value (int): Initial algorithms stage value.

        Returns:
            int, int: Current algorithms stage and maximum bits (height) allowed in a column for current stage.
        """
        stage = 0
        d = 2
        while True:
            stage += 1
            max_height = d
            # Calculating maximum height sequence
            # d(j=1) = 2; d(j+1) = floor(1.5*d)
            d = math.floor(1.5*d)
            if d >= initial_value:
                return stage, max_height

    def init_row_lengths(self):
        """Creates appropriate number of partial product rows along with filling them with corresponding number of bit pairs.

        Returns:
            list: List of partial product rows with their bit pairs.
        """
        rows = [[] for _ in range(self.N)]
        rows = [self.add_row_wires(row=row, row_index=rows.index(row)) for row in rows]
        return rows

    def add_row_wires(self, row: list, row_index: int):
        """Fills circuit's partial product row with corresponding bit pairs.

        Args:
            row (list): List representing row of partial product bits.
            row_index (int): Index of partial products row.

        Returns:
            list: Updated row list containing corresponding number of input bit pairs to form proper pp row.
        """
        # Number of partial products present in the row (should be equal to circuit's input bus size)
        row_pp_count = self.N
        # Adding neccessary number of lists (based on number of bits in the row – stored in `row_pp_count`)
        # to row that each represent individual bit pairs for described row (these bit pairs are then combined in AND/NAND gates)
        [row.append([]) for _ in range(row_pp_count)]

        # Filling row bit pair lists with appropriate bits
        [row[index].append(self.a.get_wire(index)) for index in range(row_pp_count)]
        [row[index].append(self.b.get_wire(row_index)) for index in range(row_pp_count)]

        # Converting unsigned rows of pp bit pair lists into AND gates
        if self.signed is False:
            row[0:] = [self.add_component(AndGate(a=row[i][0], b=row[i][1], prefix=self.prefix+'_and_'+str(row[i][0].index)+'_'+str(row[i][1].index), parent_component=self)).out for i in range(row_pp_count)]
        # Converting signed rows of pp bit pair lists into AND/NAND gates (based on Baugh-Wooley multiplication algorithm)
        else:
            # Partial product bit pairs of all rows (expect for the last one) are connected to AND gates, besides the last pp bit pair in each row that is connected to a NAND gate
            if row_index != self.N-1:
                row[0:row_pp_count-1] = [self.add_component(AndGate(a=row[i][0], b=row[i][1], prefix=self.prefix+'_and_'+str(row[i][0].index)+'_'+str(row[i][1].index), parent_component=self)).out for i in range(row_pp_count-1)]

                row[row_pp_count-1] = self.add_component(NandGate(a=row[row_pp_count-1][0], b=row[row_pp_count-1][1], prefix=self.prefix+'_nand_'+str(row[row_pp_count-1][0].index)+'_'+str(row[row_pp_count-1][1].index), parent_component=self)).out
            # Partial product bit pairs of the last row are connected to NAND gates besides the last pp pair that is connected to an AND gate
            else:
                row[0:row_pp_count-1] = [self.add_component(NandGate(a=row[i][0], b=row[i][1], prefix=self.prefix+'_nand_'+str(row[i][0].index)+'_'+str(row[i][1].index), parent_component=self)).out for i in range(row_pp_count-1)]

                row[row_pp_count-1] = self.add_component(AndGate(a=row[row_pp_count-1][0], b=row[row_pp_count-1][1], prefix=self.prefix+'_and_'+str(row[row_pp_count-1][0].index)+'_'+str(row[row_pp_count-1][1].index), parent_component=self)).out

        pp_row_wires = Bus(prefix=f"pp_row{row_index}", wires_list=row)
        return pp_row_wires

    def init_column_heights(self):
        """Creates appropriate number of partial product columns along with filling them with corresponding number of bit pairs.

        Returns:
            list: List of partial product columns with their bit pairs.
        """
        columns = [[num] if num <= self.N else [num - (num - self.N)*2] for num in range(1, self.out.N)]
        columns = [self.add_column_wires(column=col, column_index=columns.index(col)) for col in columns]
        return columns

    def add_column_wires(self, column: list, column_index: int):
        """Fills circuit's partial product column with corresponding bit pairs.

        Args:
            column (list): List representing column of partial product bits.
            column_index (int): Index of partial products column.

        Returns:
            list: Updated column list containing corresponding number of input bit pairs to form proper pp column.
        """
        # Adding neccessary number of lists (based on number of bits in the column – stored in `column[0]`)
        # to column that each represent individual bit pairs for described column (these bit pairs are then combined in AND/NAND gates)
        [column.append([]) for _ in range(column[0])]
        # Filling column bit pair lists with appropriate bits
        if column_index <= self.N-1:
            [column[column[0]-index].append(self.a.get_wire(index)) for index in range(0, column[0])]
            [column[index+1].append(self.b.get_wire(index)) for index in range(0, column[0])]
        else:
            [column[self.a.N-index].append(self.a.get_wire(index)) for index in range(self.a.N-1, self.a.N-column[0]-1, -1)]
            [column[index-(self.a.N-1-column[0])].append(self.b.get_wire(index)) for index in range(self.a.N-column[0], self.a.N)]

        # Converting unsigned column pp bit pair lists into AND gates
        if self.signed is False:
            column[1:] = [AndGate(a=column[i][0], b=column[i][1], prefix=self.prefix+'_and_'+str(column[i][0].index)+'_'+str(column[i][1].index), parent_component=self) for i in range(1, len(column))]
        # Converting signed column pp bit pair lists into AND/NAND gates (based on Baugh-Wooley multiplication algorithm)
        else:
            # First half of partial product columns contains only AND gates
            if column_index < self.N-1 or column_index == self.out.N-2:
                column[1:] = [AndGate(a=column[i][0], b=column[i][1], prefix=self.prefix+'_and_'+str(column[i][0].index)+'_'+str(column[i][1].index), parent_component=self) for i in range(1, len(column))]
            # Second half of partial product columns contains NAND/AND gates
            else:
                column[1] = NandGate(a=column[1][0], b=column[1][1], prefix=self.prefix+'_nand_'+str(column[1][0].index)+'_'+str(column[1][1].index), parent_component=self)
                column[-1] = NandGate(a=column[-1][0], b=column[-1][1], prefix=self.prefix+'_nand_'+str(column[-1][0].index)+'_'+str(column[-1][1].index), parent_component=self)
                if len(column[2:-1]) != 0:
                    column[2:-1] = [AndGate(a=column[i][0], b=column[i][1], prefix=self.prefix+'_and_'+str(column[i][0].index)+'_'+str(column[i][1].index), parent_component=self) for i in range(2, len(column)-1)]

        return column

    def get_column_height(self, column_num: int):
        """Retrieves the current height of desired partial products column.

        Args:
            column_num (int): Index of pp column.

        Returns:
            int: Height of the current bit column.
        """
        return self.columns[column_num][0]

    def update_column_heights(self, curr_column: int, curr_height_change: int, next_column: int = 0, next_height_change: int = 0):
        """Updates height of desired column and optionally also its subsequent column.

        Used within dadda and wallace multipliers to perform gradual reduction of partial product columns through the stages.
        Allows to choose the height change to take effect on the chosen column index and optionally also the same for the following
        column if it should also be affected.

        Args:
            curr_column (int): Current pp column index.
            curr_height_change (int): Height change for the chosen current pp column.
            next_column (int, optional): Subsequent pp column index. Defaults to 0.
            next_height_change (int, optional): Height change for the chosen subsequent pp column. Defaults to 0.
        """
        self.columns[curr_column][0] = self.get_column_height(curr_column)+curr_height_change
        if next_column-1 == curr_column:
            self.columns[next_column][0] = self.get_column_height(next_column)+next_height_change

    def add_column_wire(self, column: int, bit: int):
        """Retrieves wire from desired partial product column bit position.

        If bit pair (AND/NAND gate) is present at the desired position, it is reduced and replaced with AND/NAND gate output wire accordingly.
        Either former logic gate's output wire or present wire is returned.

        Args:
            column (int): Partial product column index.
            bit (int): Bit position within the chosen column.

        Returns:
            Wire: Return Wire present at specified position.
        """
        # Checks if a logic gate is present at desired column bit position. If so the gate is added to circuit's list of subcomponents,
        # and the former logic gates's output bit replaces the gate at desired column bit position. This output wire is also returned to the caller.
        if isinstance(self.columns[column][bit+1], AndGate) or isinstance(self.columns[column][bit+1], NandGate):
            self.add_component(self.columns[column][bit+1])
            return self.get_previous_component(1).out
        else:
            return self.columns[column][bit+1]

    def get_column_wire(self, column: int, bit: int):
        """Retrieves wire from desired partial product column bit position.

        If bit pair (AND/NAND gate) is present at the desired position, AND/NAND gate output wire is returned,
        if not the wire present at the desired position is returned.

        Args:
            column (int): Partial product column index.
            bit (int): Bit position within the chosen column.

        Returns:
            Wire: Return Wire present at specified position.
        """
        # Checks if a logic gate is present at desired column bit position. If so, its output bit is returned.
        if isinstance(self.columns[column][bit+1], AndGate) or isinstance(self.columns[column][bit+1], NandGate):
            return self.columns[column][bit+1].out
        else:
            return self.columns[column][bit+1]

    def update_column_wires(self, curr_column: int, adder: GeneralCircuit, next_column: int = 0):
        """Provides bit height reduction of the chosen column.

        Inserts chosen column's top bits into an `adder` circuit to reduce its bit height.
        Generated sum is stored to the bottom of the column and generated carry bit is stored to the top of the next column.

        Args:
            curr_column (int): Current pp column index.
            adder (GeneralCircuit): Two/three input one bit adder.
            next_column (int, optional): Subsequent pp column index. Defaults to 0.
        """
        if hasattr(adder, "c"):
            self.columns[curr_column].pop(1)
            self.columns[curr_column].pop(1)
            self.columns[curr_column].pop(1)
            self.columns[curr_column].insert(self.get_column_height(curr_column), adder.get_sum_wire())
        else:
            self.columns[curr_column].pop(1)
            self.columns[curr_column].pop(1)
            self.columns[curr_column].insert(self.get_column_height(curr_column), adder.get_sum_wire())

        if next_column-1 == curr_column:
            self.columns[next_column].insert(1, adder.get_carry_wire())



from ariths_gen.wire_components.wires import ConstantWireValue0
from ariths_gen.core.one_bit_circuits import ThreeInputOneBitCircuit
from ariths_gen.one_bit_circuits.logic_gates import AndGate, NandGate, OrGate, NorGate, XorGate, XnorGate, NotGate
from ariths_gen.wire_components import Wire, Bus


class FullAdder(ThreeInputOneBitCircuit):
    """Class representing three input one bit full adder.

    ```
        ┌──────┐
    ───►│      ├─► Sum
    ───►│      │
    ───►│      ├─► Cout
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        c (Wire, optional): Carry input wire. Defaults to Wire(name="cin").
        prefix (str, optional): Prefix name of full adder. Defaults to "".
        name (str, optional): Name of full adder. Defaults to "fa".
    """
    use_verilog_instance = False

    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), c: Wire = Wire(name="cin"), prefix: str = "", name: str = "fa"):
        super().__init__(a, b, c, prefix=prefix, name=name)
        # 2 wires for component's bus output (sum, cout)
        self.out = Bus(self.prefix+"_out", 2)

        # PG logic
        propagate_xor = XorGate(a, b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(propagate_xor)
        generate_and = AndGate(a, b, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self)
        self.add_component(generate_and)

        # Sum
        # XOR gate for calculation of 1-bit sum
        obj_xor = XorGate(propagate_xor.out, c, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=0, parent_component=self)
        self.add_component(obj_xor)
        self.out.connect(0, obj_xor.out)

        # Cout
        # AND gate for calculation of 1-bit cout
        obj_and = AndGate(propagate_xor.out, c, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self)
        self.add_component(obj_and)

        obj_or = OrGate(generate_and.out, obj_and.out, prefix=self.prefix+"_or"+str(self.get_instance_num(cls=OrGate)), outid=1, parent_component=self)
        self.add_component(obj_or)

        self.out.connect(1, obj_or.out)

    def get_sum_wire(self):
        """Get output wire carrying sum value.

        Returns:
           Wire: Return sum wire.
        """
        return self.out.get_wire(0)

    def get_carry_wire(self):
        """Get output wire carrying carry out value.

        Returns:
           Wire: Return carry out wire.
        """
        return self.out.get_wire(1)

    def get_init_v_flat(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_init_v_flat()

        return "  " + self.use_verilog_instance.format(
            **{
                "unit": self.prefix,
                "wirea": f"1'b{self.a.value}" if self.a.is_const() else self.a.name,
                "wireb": f"1'b{self.b.value}" if self.b.is_const() else self.b.name,
                "wirec": f"1'b{self.c.value}" if self.c.is_const() else self.c.name,
                "wireys": self.get_sum_wire().prefix,
                "wireyc": self.get_carry_wire().prefix,
            }) + ";\n"

    def get_self_init_v_hier(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_self_init_v_hier()

        unique_out_wires = []
        for o in self.out.bus:
            unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name)

        return "  " + self.use_verilog_instance.format(
            **{
                "unit": self.prefix,
                "wirea": self.a.name,
                "wireb": self.b.name,
                "wirec": self.c.name,
                "wireys": unique_out_wires[0],
                "wireyc": unique_out_wires[1],
            }) + ";\n"

    def get_circuit_v(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_circuit_v()

        return f"{self.get_prototype_v_hier()}" + \
               f"{self.get_self_init_v_hier()}" + \
               f"endmodule"


class FullAdderP(FullAdder, ThreeInputOneBitCircuit):
    """Class representing three input one bit full adder with additional output wire for P signal.

    ```
        ┌──────┐
    ───►│      ├─► Sum
    ───►│      ├─► Cout
    ───►│      ├─► P
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        c (Wire, optional): Carry input wire. Defaults to Wire(name="cin").
        prefix (str, optional): Prefix name of full adder with p logic. Defaults to "".
        name (str, optional): Name of full adder with p logic. Defaults to "fa_p".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), c: Wire = Wire(name="cin"), prefix: str = "", name: str = "fa_p"):
        super().__init__(a, b, c, prefix=prefix, name=name)
        # 3 wires for component's bus output (sum, cout, propagate)
        self.out.bus_extend(3)
        self.out.connect(2, self.get_previous_component(5).out)

    def get_propagate_wire(self):
        """Get output wire carrying propagate value.

        Returns:
            Wire: Return propagate wire.
        """
        return self.out.get_wire(2)


class FullAdderPG(FullAdder, ThreeInputOneBitCircuit):
    """Class representing three input one bit full adder with additional output wires for P and G signals.

    ```
        ┌──────┐
    ───►│      ├─► Sum
    ───►│      ├─► Cout
    ───►│      ├─► P
        │      ├─► G
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        c (Wire, optional): Carry input wire. Defaults to Wire(name="cin").
        prefix (str, optional): Prefix name of full adder with pg logic. Defaults to "".
        name (str, optional): Name of full adder with pg logic. Defaults to "fa_pg".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), c: Wire = Wire(name="cin"), prefix: str = "", name: str = "fa_pg"):
        super().__init__(a, b, c, prefix=prefix, name=name)
        # 4 wires for component's bus output (sum, cout, propagate, generate)
        self.out.bus_extend(4)
        self.out.connect(2, self.get_previous_component(5).out)
        self.out.connect(3, self.get_previous_component(4).out)

    def get_propagate_wire(self):
        """Get output wire carrying propagate value.

        Returns:
            Wire: Return propagate wire.
        """
        return self.out.get_wire(2)

    def get_generate_wire(self):
        """Get output wire carrying generate value.

        Returns:
            Wire: Return generate wire.
        """
        return self.out.get_wire(3)


class PGSumLogic(ThreeInputOneBitCircuit):
    """Class represents a three input function block that contains logic for obtaining the propagate/generate/sum signals.

    ```
        ┌──────┐
    ───►│      ├─► P
    ───►│      ├─► G
    ───►│      ├─► Sum
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        c (Wire, optional): Carry input wire. Defaults to Wire(name="cin").
        prefix (str, optional): Prefix name of pg sum logic. Defaults to "".
        name (str, optional): Name of pg sum logic. Defaults to "pg_sum".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), c: Wire = Wire(name="cin"), prefix: str = "", name: str = "pg_sum"):
        super().__init__(a, b, c, prefix=prefix, name=name)
        # 3 wires for component's bus output (sum, propagate, generate)
        self.out = Bus(self.prefix+"_out", 3)

        # PG logic
        propagate_xor = XorGate(a, b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=0, parent_component=self)
        self.add_component(propagate_xor)
        self.out.connect(0, propagate_xor.out)

        generate_and = AndGate(a, b, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), outid=1, parent_component=self)
        self.add_component(generate_and)
        self.out.connect(1, generate_and.out)

        # Sum output
        sum_xor = XorGate(propagate_xor.out, c, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=2, parent_component=self)
        self.add_component(sum_xor)
        self.out.connect(2, sum_xor.out)

    def get_propagate_wire(self):
        """Get output wire carrying propagate signal value.

        Returns:
           Wire: Return propagate wire.
        """
        return self.out.get_wire(0)

    def get_generate_wire(self):
        """Get output wire carrying generate signal value.

        Returns:
           Wire: Return generate wire.
        """
        return self.out.get_wire(1)

    def get_sum_wire(self):
        """Get output wire carrying sum value.

        Returns:
           Wire: Return sum wire.
        """
        return self.out.get_wire(2)


class TwoOneMultiplexer(ThreeInputOneBitCircuit):
    """Class representing two to one multiplexer (with select signal as its third input).

    ```
        ┌──────┐
    ───►│      │
        │      ├─►
    ───►│      │
        └───▲──┘
    Sel ────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="d0").
        b (Wire, optional): Second input wire. Defaults to Wire(name="d1").
        c (Wire, optional): Select signal. Defaults to Wire(name="sel").
        prefix (str, optional): Prefix name of two to one multiplexer. Defaults to "".
        name (str, optional): Name of two to one multiplexer. Defaults to "mux2to1".
    """
    use_verilog_instance = False

    def __init__(self, a: Wire = Wire(name="d0"), b: Wire = Wire(name="d1"), c: Wire = Wire(name="sel"), prefix: str = "", name: str = "mux2to1"):
        super().__init__(a, b, c, prefix=prefix, name=name)
        # Represents select signal (self.c naming for proper unified generation)
        self.c = c
        # 1 wire for component's output bus
        self.out = Bus(self.prefix+"_out", 1)

        # 2:1MUX logic
        and_obj = AndGate(a=self.b, b=self.c, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self)
        self.add_component(and_obj)

        not_obj = NotGate(a=self.c, prefix=self.prefix+"_not"+str(self.get_instance_num(cls=NotGate)), parent_component=self)
        self.add_component(not_obj)

        and_obj = AndGate(a=self.a, b=self.get_previous_component().out, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self)
        self.add_component(and_obj)

        xor_obj = XorGate(a=self.get_previous_component(3).out, b=self.get_previous_component().out, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(xor_obj)

        # Connection of MUX output wire
        self.out.connect(0, xor_obj.out)

    def get_mux_out_wire(self):
        """Get multiplexer output wire.

        Returns:
           Wire: Return multiplexer out wire.
        """
        return self.out.get_wire(0)

    def get_init_v_flat(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_init_v_flat()

        if self.out[0].is_const():
            return ""
        else:
            # TODO - replace by one verilog_instance_format!
            neg_out_w_name = f"neg_{self.out.get_wire(0).name}"
            return f"  wire {neg_out_w_name};\n  " + self.use_verilog_instance.format(
                **{
                    "unit": self.prefix,
                    "wirea": self.a.get_wire_value_v_hier(), # former version:  f"1'b{self.a.value}" if self.a.is_const() else self.a.name,
                    "wireb": self.b.get_wire_value_v_hier(), #f"1'b{self.b.value}" if self.b.is_const() else self.b.name,
                    "wires": self.c.get_wire_value_v_hier(), #f"1'b{self.c.value}" if self.c.is_const() else self.c.name,
                    "wirey": neg_out_w_name,
                }) + ";\n" + f"  assign {self.out.get_wire(0).name} = ~{neg_out_w_name};\n"

    def get_self_init_v_hier(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_self_init_v_hier()

        unique_out_wires = []
        for o in self.out.bus:
            unique_out_wires.append(o.name+"_outid"+str(self.out.bus.index(o))) if o.is_const() or o.name in [self.a.name, self.b.name] else unique_out_wires.append(o.name)

        neg_out_w_name = f"neg_{unique_out_wires[0]}"
        return f"  wire {neg_out_w_name};\n  " + self.use_verilog_instance.format(
            **{
                "unit": self.prefix,
                "wirea": self.a.name,
                "wireb": self.b.name,
                "wires": self.c.name,
                "wirey": neg_out_w_name
            }) + ";\n" + f"  assign {unique_out_wires[0]} = ~{neg_out_w_name};\n"

    def get_circuit_v(self):
        """ support of custom PDK """
        if not self.use_verilog_instance:
            return super().get_circuit_v()

        return f"{self.get_prototype_v_hier()}" + \
               f"{self.get_self_init_v_hier()}" + \
               f"endmodule"


class FullSubtractor(ThreeInputOneBitCircuit):
    """Class representing three input one bit full subtractor.

    ```
        ┌──────┐
    ───►│      ├─► Difference
    ───►│      │
    ───►│      ├─► Bout
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire. Defaults to Wire(name="a").
        b (Wire, optional): Second input wire. Defaults to Wire(name="b").
        c (Wire, optional): Input borrow wire. Defaults to Wire(name="bin").
        prefix (str, optional): Prefix name of full subtractor. Defaults to "".
        name (str, optional): Name of full subtractor. Defaults to "fs".
    """
    def __init__(self, a: Wire = Wire(name="a"), b: Wire = Wire(name="b"), c: Wire = Wire(name="bin"), prefix: str = "", name: str = "fs"):
        super().__init__(a, b, c, prefix=prefix, name=name)
        # 2 wires for component's bus output (difference, bout)
        self.out = Bus(self.prefix+"_out", 2)

        # Difference
        xor_obj = XorGate(a=self.a, b=self.b, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(xor_obj)

        not_obj = NotGate(a=self.a, prefix=self.prefix+"_not"+str(self.get_instance_num(cls=NotGate)), parent_component=self)
        self.add_component(not_obj)

        and_obj = AndGate(a=not_obj.out, b=self.b, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self)
        self.add_component(and_obj)

        difference_xor = XorGate(a=self.c, b=xor_obj.out, prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), outid=0, parent_component=self)
        self.add_component(difference_xor)
        self.out.connect(0, difference_xor.out)

        # Borrow out
        not_obj = NotGate(a=xor_obj.out, prefix=self.prefix+"_not"+str(self.get_instance_num(cls=NotGate)), parent_component=self)
        self.add_component(not_obj)

        and_obj = AndGate(a=not_obj.out, b=self.c, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self)
        self.add_component(and_obj)

        borrow_out_or = OrGate(a=and_obj.out, b=self.get_previous_component(4).out, prefix=self.prefix+"_or"+str(self.get_instance_num(cls=OrGate)), outid=1, parent_component=self)
        self.add_component(borrow_out_or)
        self.out.connect(1, borrow_out_or.out)

    def get_difference_wire(self):
        """Get output wire carrying difference value.

        Returns:
           Wire: Return difference wire.
        """
        return self.out.get_wire(0)

    def get_borrow_wire(self):
        """Get output wire carrying borrow out value.

        Returns:
           Wire: Return borrow out wire.
        """
        return self.out.get_wire(1)


class GreyCell(ThreeInputOneBitCircuit):
    """Class representing three input grey cell used in parallel prefix adders inside the PG (parallel prefix computation) logic.

    Grey cell is based on a notation used in the book CMOS VLSI Design (please refer to the book for more details not deducible from the code itself).
    ```
        ┌──────┐
    ───►│      │
    ───►│      ├─► Generate
    ───►│      │
        └──────┘
    ```

    Description of the __init__ method.

    Args:
        a (Wire, optional): First input wire, represents generate signal from the current stage. Defaults to Wire(name="g1").
        b (Wire, optional): Second input wire, represents propagate signal from the current stage. Defaults to Wire(name="p1").
        c (Wire, optional): Third input wire, represents generate signal from a preceding stage. Defaults to Wire(name="g0").
        prefix (str, optional): Prefix name of grey cell. Defaults to "".
        name (str, optional): Name of grey cell. Defaults to "gc".
    """
    def __init__(self, a: Wire = Wire(name="g1"), b: Wire = Wire(name="p1"), c: Wire = Wire(name="g0"), prefix: str = "", name: str = "gc"):
        super().__init__(a, b, c, prefix=prefix, name=name)
        # 1 wire for component's bus output (output generate)
        self.out = Bus(self.prefix+"_out", 1)

        # Create/propagate the generate signal
        self.add_component(AndGate(a=self.b, b=self.c, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self))
        self.add_component(OrGate(a=self.a, b=self.get_previous_component().out, prefix=self.prefix+"_or"+str(self.get_instance_num(cls=OrGate)), parent_component=self))
        # Connection of the generate output wire
        self.out.connect(0, self.get_previous_component().out)

    def get_generate_wire(self):
        """Get output wire carrying generate value.

        Returns:
           Wire: Return generate wire.
        """
        return self.out.get_wire(0)



from ariths_gen.multi_bit_circuits.multipliers.array_multiplier import (
    UnsignedArrayMultiplier,
    SignedArrayMultiplier
)

from ariths_gen.multi_bit_circuits.multipliers.wallace_multiplier import (
    UnsignedWallaceMultiplier,
    SignedWallaceMultiplier
)

from ariths_gen.multi_bit_circuits.multipliers.dadda_multiplier import (
    UnsignedDaddaMultiplier,
    SignedDaddaMultiplier
)

from ariths_gen.multi_bit_circuits.multipliers.carry_save_multiplier import (
    UnsignedCarrySaveMultiplier,
    SignedCarrySaveMultiplier
)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    XnorGate,
    NotGate
)
from ariths_gen.multi_bit_circuits.others import (
    OrReduce
)


class UnsignedCompareLT(GeneralCircuit):
    """Class representing unsigned compare

    
    Returns true if a < b
    """

    def __init__(self, a: Bus, b: Bus, prefix : str = "", name : str = "cmp_lt", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(name=name, prefix=prefix, inputs=[a, b], out_N=1)

        # create wires
        psum = ConstantWireValue1()

        res = Bus(N = self.N, prefix=self.prefix + "res")

        for i in reversed(range(self.N)):
            iA = self.a[i] if i < self.a.N else ConstantWireValue0()
            iB = self.b[i] if i < self.b.N else ConstantWireValue0()
            
            i1 = self.add_component(NotGate(iA, f"{self.prefix}_i1_{i}")).out
            i2 = iB

            and1 = self.add_component(AndGate(i1, i2, f"{self.prefix}_and1_{i}")).out
            res[i] = self.add_component(AndGate(and1, psum, f"{self.prefix}_and2_{i}")).out

            pi = self.add_component(XnorGate(iA, iB, f"{self.prefix}_pi_{i}", parent_component=self)).out
            psum = self.add_component(AndGate(pi, psum, f"{self.prefix}_psum_{i}")).out


        red = self.add_component(OrReduce(res, prefix=f"{self.prefix}_orred", inner_component=True))
        self.out.connect_bus(red.out)


class UnsignedCompareLTE(GeneralCircuit):
    """Class representing unsigned compare
    
    Returns true if a <= b
    """

    def __init__(self, a: Bus, b: Bus, prefix : str = "", name : str = "cmp_lte", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(name=name, prefix=prefix, inputs=[a, b], out_N=1)

        # create wires
        psum = ConstantWireValue1()

        res = Bus(N = self.N + 1, prefix=self.prefix + "res")

        for i in reversed(range(self.N)):
            iA = self.a[i] if i < self.a.N else ConstantWireValue0()
            iB = self.b[i] if i < self.b.N else ConstantWireValue0()
            
            i1 = self.add_component(NotGate(iA, f"{self.prefix}_i1_{i}")).out
            i2 = iB

            and1 = self.add_component(AndGate(i1, i2, f"{self.prefix}_and1_{i}")).out
            res[i] = self.add_component(AndGate(and1, psum, f"{self.prefix}_and2_{i}")).out

            pi = self.add_component(XnorGate(iA, iB, f"{self.prefix}_pi_{i}", parent_component=self)).out
            psum = self.add_component(AndGate(pi, psum, f"{self.prefix}_psum_{i}")).out

        res[self.N] = psum # or all equal (xor)

        red = self.add_component(OrReduce(res, prefix=f"{self.prefix}_orred", inner_component=True))

        self.out.connect_bus(red.out)



class UnsignedCompareGT(GeneralCircuit):
    """Class representing unsigned compare

    
    Returns true if a < b
    """

    def __init__(self, a: Bus, b: Bus, prefix : str = "", name : str = "cmp_gt", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(name=name, prefix=prefix, inputs=[a, b], out_N=1)

        # create wires
        psum = ConstantWireValue1()

        res = Bus(N = self.N, prefix=self.prefix + "res")

        for i in reversed(range(self.N)):
            iA = self.a[i] if i < self.a.N else ConstantWireValue0()
            iB = self.b[i] if i < self.b.N else ConstantWireValue0()
            
            i1 = iA
            i2 = self.add_component(NotGate(iB, f"{self.prefix}_i2_{i}")).out

            and1 = self.add_component(AndGate(i1, i2, f"{self.prefix}_and1_{i}")).out
            res[i] = self.add_component(AndGate(and1, psum, f"{self.prefix}_and2_{i}")).out

            pi = self.add_component(XnorGate(iA, iB, f"{self.prefix}_pi_{i}", parent_component=self)).out
            psum = self.add_component(AndGate(pi, psum, f"{self.prefix}_psum_{i}")).out


        red = self.add_component(OrReduce(res, prefix=f"{self.prefix}_orred", inner_component=True))
        self.out.connect_bus(red.out)


class UnsignedCompareGTE(GeneralCircuit):
    """Class representing unsigned compare
    
    Returns true if a <= b
    """

    def __init__(self, a: Bus, b: Bus, prefix : str = "", name : str = "cmp_gte", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(name=name, prefix=prefix, inputs=[a, b], out_N=1)

        # create wires
        psum = ConstantWireValue1()

        res = Bus(N = self.N + 1, prefix=self.prefix + "res")

        for i in reversed(range(self.N)):
            iA = self.a[i] if i < self.a.N else ConstantWireValue0()
            iB = self.b[i] if i < self.b.N else ConstantWireValue0()

            i1 = iA
            i2 = self.add_component(NotGate(iB, f"{self.prefix}_i1_{i}", parent_component=self)).out

            and1 = self.add_component(AndGate(i1, i2, f"{self.prefix}_and1_{i}", parent_component=self)).out
            res[i] = self.add_component(AndGate(and1, psum, f"{self.prefix}_and2_{i}", parent_component=self)).out

            pi = self.add_component(XnorGate(iA, iB, f"{self.prefix}_pi_{i}", parent_component=self)).out
            psum = self.add_component(AndGate(pi, psum, f"{self.prefix}_psum_{i}", parent_component=self)).out

        res[self.N] = psum # or all equal (xor)
        
        red = self.add_component(OrReduce(res, prefix=f"{self.prefix}_orred", inner_component=True, parent_component=self))

        self.out.connect_bus(red.out)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate
)


class UnsignedBrokenArrayMultiplier(MultiplierCircuit):
    """Class representing unsigned broken array multiplier.

    It represents an approximative version of unsigned array multiplier with simpler structure.
    It is created by modifying an ordinary N-bit unsigned array multiplier by omitting partial product
    stage cells by the specified horizontal and vertical cut levels.

    The design promises better area and power parameters in exchange for the loss of computation precision.
    The BAM design allows to save more partial product stage adders than truncated multiplier architectures.
    ```
                            VERTICAL CUT=4

                                │
                                            A3B0       A2B0       A1B0       A0B0
                                │          ┌───┐      ┌───┐      ┌───┐      ┌───┐
                                           │AND│      │AND│      │AND│      │AND│
                                │          └───┘      └───┘      └───┘      └───┘

                                │
                                      A3B1       A2B1       A1B1       A0B1
                                │    ┌───┐      ┌───┐      ┌───┐      ┌───┐
                                     │AND│      │AND│      │AND│      │AND│
                                │    └───┘      └───┘      └───┘      └───┘

                                │ ┌────┐     ┌────┐     ┌────┐     ┌────┐
                                  │    │     │    │     │    │     │    │
                                │ │ HA │     │ FA │     │ FA │     │ HA │
                                  │    │     │    │     │    │     │    │
                                │ └────┘     └────┘     └────┘     └────┘

              ─ ─ ─ ─ ─ ─ ─ ─ ─ ┼ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ HORIZONTAL CUT=2
                       A3B2           A2B2       A1B2       A0B2
                      ┌▼─▼┐     │    ┌───┐      ┌───┐      ┌───┐
                      │AND│          │AND│      │AND│      │AND│
                      └┬──┘     │    └───┘      └───┘      └───┘
                       │
                       │        │ ┌────┐     ┌────┐     ┌────┐
                       │          │    │     │    │     │    │
                       │        │ │ FA │     │ FA │     │ HA │
                       │          │    │     │    │     │    │
                       │        │ └────┘     └────┘     └────┘
            A3B3       │  A2B3       A1B3      A0B3
           ┌▼─▼┐       │ ┌▼─▼┐  │   ┌───┐     ┌───┐
           │AND│       │ │AND│      │AND│     │AND│
           └┬──┘       │ └┬──┘  │   └───┘     └───┘
            │          │  │
           ┌▼───┐     ┌▼──▼┐    │ ┌────┐    ┌────┐
           │    │     │    │      │    │    │    │
    ┌──────┤ HA │◄────┤ HA │    │ │ FA │    │ HA │
    │      │    │     │    │      │    │    │    │
    │      └──┬─┘     └──┬─┘    │ └────┘    └────┘
    │         │          │
    │         │          │      │
    ▼         ▼          ▼          ▼           ▼          ▼          ▼          ▼
    P7        P6         P5     │   P4=0        P3=0       P2=0       P1=0       P0=0
    ```
    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        horizontal_cut (int, optional): Specifies horizontal cut used in broken array multiplier circuit creation. Defaults to 0.
        vertical_cut (int, optional): Specifies vertical cut used in broken array multiplier circuit creation. Defaults to 0.
        prefix (str, optional): Prefix name of unsigned broken array multiplier. Defaults to "".
        name (str, optional): Name of unsigned broken array multiplier. Defaults to "u_bam".
    """
    def __init__(self, a: Bus, b: Bus, horizontal_cut: int = 0, vertical_cut: int = 0, prefix: str = "", name: str = "u_bam", **kwargs):
        # NOTE: If horizontal/vertical cut is specified as 0 the final circuit is a simple array multiplier
        self.horizontal_cut = horizontal_cut
        self.vertical_cut = vertical_cut

        self.N = max(a.N, b.N)
        # Horizontal cut level should be: 0 <= horizontal_cut < N
        # Vertical cut level should be: horizontal_cut <= vertical_cut < 2*N
        assert horizontal_cut < self.N
        assert vertical_cut < 2*self.N

        # Vertical cut should be greater or equal to horizontal cut
        assert vertical_cut >= horizontal_cut

        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        self.ommited_rows = 0
        # Gradual generation of partial products
        for b_multiplier_index in range(self.horizontal_cut, self.N):
            # Number of elements that should be ommited in the current level based on vertical cut
            pp_row_elems_to_skip = self.vertical_cut - b_multiplier_index if self.vertical_cut - b_multiplier_index > 0 else 0
            # Number of pp pairs present in the current row
            pp_row_elems = self.N-pp_row_elems_to_skip if self.N-pp_row_elems_to_skip > 0 else 0
            self.ommited_rows += 1 if pp_row_elems == 0 else 0

            for a_multiplicand_index in range((self.N-pp_row_elems), self.N):
                # AND gates generation for calculation of partial products
                obj_and = AndGate(self.a.get_wire(a_multiplicand_index), self.b.get_wire(b_multiplier_index), prefix=self.prefix+"_and"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                self.add_component(obj_and)

                if b_multiplier_index != self.horizontal_cut + self.ommited_rows:
                    if b_multiplier_index == self.horizontal_cut + self.ommited_rows + 1:
                        previous_product = self.components[a_multiplicand_index + b_multiplier_index - self.vertical_cut].out
                    else:
                        previous_product = self.get_previous_partial_product(a_index=a_multiplicand_index, b_index=b_multiplier_index, mult_type="bam")

                    # HA generation for first 1-bit adder in each row starting from the second one
                    if a_multiplicand_index == 0 or self.vertical_cut-b_multiplier_index == a_multiplicand_index:
                        obj_adder = HalfAdder(self.get_previous_component().out, previous_product, prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)
                        # Product generation
                        self.out.connect(b_multiplier_index, obj_adder.get_sum_wire())

                    # HA generation, last 1-bit adder in second row
                    elif a_multiplicand_index == self.N-1 and b_multiplier_index == self.horizontal_cut+self.ommited_rows+1:
                        obj_adder = HalfAdder(self.get_previous_component().out, self.get_previous_component(number=2).get_carry_wire(), prefix=self.prefix+"_ha"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)

                    # FA generation
                    else:
                        obj_adder = FullAdder(self.get_previous_component().out, previous_product, self.get_previous_component(number=2).get_carry_wire(), prefix=self.prefix+"_fa"+str(a_multiplicand_index)+"_"+str(b_multiplier_index))
                        self.add_component(obj_adder)

                # PRODUCT GENERATION
                if (a_multiplicand_index == 0 and b_multiplier_index == self.horizontal_cut) or (self.horizontal_cut + self.ommited_rows == self.N-1):
                    self.out.connect(a_multiplicand_index + b_multiplier_index, obj_and.out)

                    # 1 bit multiplier case
                    if a_multiplicand_index == self.N-1 and b_multiplier_index == self.N-1:
                        self.out.connect(a_multiplicand_index+b_multiplier_index+1, ConstantWireValue0())

                elif b_multiplier_index == self.N-1 and self.horizontal_cut != self.N-1:
                    self.out.connect(b_multiplier_index + a_multiplicand_index, obj_adder.get_sum_wire())

                    if a_multiplicand_index == self.N-1:
                        self.out.connect(self.out.N-1, obj_adder.get_carry_wire())

        # Connecting the output bits generated from ommited cells to ground
        if self.vertical_cut == 2*self.N-1:
            [self.out.connect(out_id, ConstantWireValue0()) for out_id in range(self.out.N)]
        else:
            for grounded_out_index in range(0, max(self.horizontal_cut, self.vertical_cut)):
                self.out.connect(grounded_out_index, ConstantWireValue0())



from ariths_gen.multi_bit_circuits.adders.carry_lookahead_adder import (
    UnsignedCarryLookaheadAdder
)
from ariths_gen.wire_components import (
    ConstantWireValue0,
    ConstantWireValue1,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    MultiplierCircuit
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder
)
from ariths_gen.one_bit_circuits.logic_gates import (
    XorGate
)


class UnsignedDaddaMultiplier(MultiplierCircuit):
    """Class representing unsigned dadda multiplier.

    Unsigned dadda multiplier represents fast N-bit multiplier which utilizes
    the functionality of reduction algorithm proposed by Luigi Dadda.

    First partial products are calculated for each bit pair that form the partial product multiplication columns.
    At last the reduced pairs are inserted into chosen multi bit unsigned adder to execute their summation and obtain the final output bits.

    Dadda algorithm is described more in detail here:
    https://en.wikipedia.org/wiki/Dadda_multiplier

    It is composed of much less inner components (half/full adders, AND gates) as opposed
    to e.g. wallace and array multipliers.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of unsigned dadda multiplier. Defaults to "".
        name (str, optional): Name of unsigned dadda multiplier. Defaults to "u_dadda_cla".
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used to obtain final sums of products. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "u_dadda_cla", unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        # Get starting stage and maximum possible column height
        self.stage, self.d = self.get_maximum_height(initial_value=min(self.a.N, self.b.N))
        # Initialize all columns partial products forming AND gates matrix
        self.columns = self.init_column_heights()

        # Perform reduction until stage 0
        for stage in range(self.stage, 0, -1):
            col = 0
            while col < len(self.columns):
                if self.get_column_height(col) == self.d + 1:
                    # Add half adder and also AND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                    obj_adder = HalfAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                    self.add_component(obj_adder)

                    # Update the number of current and next column wires
                    self.update_column_heights(curr_column=col, curr_height_change=-1, next_column=col+1, next_height_change=1)

                    # Update current and next column wires arrangement
                    #   add ha's generated sum to the bottom of current column
                    #   add ha's generated cout to the top of next column
                    self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))

                elif self.get_column_height(col) > self.d:
                    # Add full adder and also AND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                    obj_adder = FullAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), self.add_column_wire(column=col, bit=2), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))
                    self.add_component(obj_adder)

                    # Update the number of current and next column wires
                    self.update_column_heights(curr_column=col, curr_height_change=-2, next_column=col+1, next_height_change=1)

                    # Update current and next column wires arrangement
                    #   add fa's generated sum to the bottom of current column
                    #   add fa's generated cout to the top of next column
                    self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))

                    # Next iteration with same column in case there is need for further reduction
                    col -= 1
                col += 1
            # Update maximum possible column height
            _, self.d = self.get_maximum_height(stage)

        # Output generation
        # First output bit from single first pp AND gate
        self.out.connect(0, self.add_column_wire(column=0, bit=0))
        # Final addition of remaining bits
        # 1 bit multiplier case
        if self.N == 1:
            self.out.connect(1, ConstantWireValue0())
        # 2 bit multiplier case
        elif self.N == 2:
            obj_ha = HalfAdder(self.add_column_wire(column=1, bit=0), self.add_column_wire(column=1, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
            self.add_component(obj_ha)
            self.out.connect(1, obj_ha.get_sum_wire())

            obj_ha = HalfAdder(self.get_previous_component().get_carry_wire(), self.add_column_wire(column=2, bit=0), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
            self.add_component(obj_ha)
            self.out.connect(2, obj_ha.get_sum_wire())
            self.out.connect(3, obj_ha.get_carry_wire())
        # Final addition of remaining bits using chosen unsigned multi bit adder
        else:
            # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
            adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(len(self.columns)-1)
            adder_a = Bus(prefix=f"a", wires_list=[self.add_column_wire(column=col, bit=0) for col in range(1, len(self.columns))])
            adder_b = Bus(prefix=f"b", wires_list=[self.add_column_wire(column=col, bit=1) for col in range(1, len(self.columns))])
            final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
            self.add_component(final_adder)

            [self.out.connect(o, final_adder.out.get_wire(o-1), inserted_wire_desired_index=o-1) for o in range(1, len(self.out.bus))]


class SignedDaddaMultiplier(MultiplierCircuit):
    """Class representing signed dadda multiplier.

    Signed dadda multiplier represents fast N-bit multiplier which utilizes
    the functionality of reduction algorithm proposed by Luigi Dadda and uses Baugh-Wooley algorithm
    to perform signed multiplication.

    First partial products are calculated for each bit pair that form the partial product multiplication columns.
    At last the reduced pairs are inserted into chosen multi bit unsigned adder to execute their summation and obtain the final output bits,
    additional XOR gate serve the necessary sign extension.

    Dadda algorithm is described more in detail here:
    https://en.wikipedia.org/wiki/Dadda_multiplier

    It is composed of much less inner components (half/full adders, AND/NAND gates) as opposed
    to e.g. wallace and array multipliers.

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        prefix (str, optional): Prefix name of signed dadda multiplier. Defaults to "".
        name (str, optional): Name of signed dadda multiplier. Defaults to "s_dadda_cla".
        unsigned_adder_class_name (str, optional): Unsigned multi bit adder used to obtain final sums of products. Defaults to UnsignedCarryLookaheadAdder.
    """
    def __init__(self, a: Bus, b: Bus, prefix: str = "", name: str = "s_dadda_cla", unsigned_adder_class_name: str = UnsignedCarryLookaheadAdder, **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N*2, signed=True, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix, desired_extension_wire=self.a.get_wire(self.a.N-1))
        self.b.bus_extend(N=self.N, prefix=b.prefix, desired_extension_wire=self.b.get_wire(self.b.N-1))

        # Get starting stage and maximum possible column height
        self.stage, self.d = self.get_maximum_height(initial_value=min(self.a.N, self.b.N))
        # Initialize all columns partial products forming AND/NAND gates matrix based on Baugh-Wooley multiplication
        self.columns = self.init_column_heights()

        # Not used for 1 bit multiplier
        if self.N != 1:
            # Adding constant wire with value 1 to achieve signedness based on Baugh-Wooley multiplication algorithm
            # (adding constant value bit to last column (with one bit) to combine them in XOR gate to get the correct final multplication output bit at the end)
            self.columns[self.N].insert(1, ConstantWireValue1())
            self.update_column_heights(curr_column=self.N, curr_height_change=1)

        # Perform reduction until stage 0
        for stage in range(self.stage, 0, -1):
            col = 0
            while col < len(self.columns):
                if self.get_column_height(col) == self.d + 1:
                    # Add half adder and also AND/NAND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                    obj_adder = HalfAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
                    self.add_component(obj_adder)

                    # Update the number of current and next column wires
                    self.update_column_heights(curr_column=col, curr_height_change=-1, next_column=col+1, next_height_change=1)

                    # Update current and next column wires arrangement
                    #   add ha's generated sum to the bottom of current column
                    #   add ha's generated cout to the top of next column
                    self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))
                elif self.get_column_height(col) > self.d:
                    # Add full adder and also AND/NAND gates if neccesarry (via add_column_wire invocation) into list of circuit components
                    obj_adder = FullAdder(self.add_column_wire(column=col, bit=0), self.add_column_wire(column=col, bit=1), self.add_column_wire(column=col, bit=2), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))
                    self.add_component(obj_adder)

                    # Update the number of current and next column wires
                    self.update_column_heights(curr_column=col, curr_height_change=-2, next_column=col+1, next_height_change=1)

                    # Update current and next column wires arrangement
                    #   add fa's generated sum to the bottom of current column
                    #   add fa's generated cout to the top of next column
                    self.update_column_wires(curr_column=col, next_column=col+1, adder=self.get_previous_component(1))
                    # Next iteration with same column in case there is need for further reduction
                    col -= 1
                col += 1
            # Update maximum possible column height
            _, self.d = self.get_maximum_height(stage)

        # Output generation
        # First output bit from single first pp AND gate
        self.out.connect(0, self.add_column_wire(column=0, bit=0))
        # Final addition of remaining bits
        # 1 bit multiplier case (no sign extension)
        if self.N == 1:
            self.out.connect(1, ConstantWireValue0())
            return
        # 2 bit multiplier case
        elif self.N == 2:
            obj_ha = HalfAdder(self.add_column_wire(column=1, bit=0), self.add_column_wire(column=1, bit=1), prefix=self.prefix+"_ha"+str(self.get_instance_num(cls=HalfAdder)))
            self.add_component(obj_ha)
            self.out.connect(1, obj_ha.get_sum_wire())

            obj_fa = FullAdder(self.get_previous_component().get_carry_wire(), self.add_column_wire(column=2, bit=0), self.add_column_wire(column=2, bit=1), prefix=self.prefix+"_fa"+str(self.get_instance_num(cls=FullAdder)))
            self.add_component(obj_fa)
            self.out.connect(2, obj_fa.get_sum_wire())
            self.out.connect(3, obj_fa.get_carry_wire())

        # Final addition of remaining bits using chosen unsigned multi bit adder
        else:
            # Obtain proper adder name with its bit width (columns bit pairs minus the first alone bit)
            adder_name = unsigned_adder_class_name(a=a, b=b).prefix + str(len(self.columns)-1)
            adder_a = Bus(prefix=f"a", wires_list=[self.add_column_wire(column=col, bit=0) for col in range(1, len(self.columns))])
            adder_b = Bus(prefix=f"b", wires_list=[self.add_column_wire(column=col, bit=1) for col in range(1, len(self.columns))])
            final_adder = unsigned_adder_class_name(a=adder_a, b=adder_b, prefix=self.prefix, name=adder_name, inner_component=True, **kwargs)
            self.add_component(final_adder)

            [self.out.connect(o, final_adder.out.get_wire(o-1), inserted_wire_desired_index=o-1) for o in range(1, len(self.out.bus))]

        # Final XOR to ensure proper sign extension
        obj_xor = XorGate(ConstantWireValue1(), self.out.get_wire(self.out.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate)), parent_component=self)
        self.add_component(obj_xor)
        self.out.connect(self.out.N-1, obj_xor.out)



from .two_input_one_bit_circuit import (
    TwoInputOneBitCircuit
)

from .three_input_one_bit_circuit import (
    ThreeInputOneBitCircuit
)

from .four_input_one_bit_circuit import (
    FourInputOneBitCircuit
)



from ariths_gen.wire_components import (
    ConstantWireValue0,
    Bus
)
from ariths_gen.core.arithmetic_circuits import (
    GeneralCircuit
)
from ariths_gen.core.logic_gate_circuits import (
    MultipleInputLogicGate
)
from ariths_gen.one_bit_circuits.one_bit_components import (
    HalfAdder,
    FullAdder,
    FullAdderP
)
from ariths_gen.one_bit_circuits.logic_gates import (
    AndGate,
    OrGate,
    XorGate
)


class UnsignedCarryIncrementAdder(GeneralCircuit):
    """Class representing unsigned carry increment adder.

    Carry increment adder represents a modified carry select adder that achieves about the same critical delay
    while containing less circuitry. It achieves this by factoring out the common logic from the two
    chains of full adders (one with cin of 0, the other with 1) present in a classical carry select adder and
    by replacing multiplexers with HAs.

    Each carry increment block is composed of these logic parts:
    Full adder chain with cin 0 is used to obtain intermediate sum, propagate and cout signals. (In case of first block actual sum bits are obtained)
    From second block onward the blocks also contain a chain of half adders to calculate the final sum bits.
    Finally the intermediate propagate signals are ANDed with the cout of the half adder chain and ORed with the intermediate
    full adder cout to determine the carry bit for the next increment block.

    ```
                           ┼   ┼                ┼   ┼
                   Cout┌───▼───▼───┐            │   │
       ┌───────────────┤    FAs    │◄─0         │   │
       │               └┬────────┬─┘            │   │
       │    ┌─────┐     ┼ P sigs ┼ Sums         │   │
       │    │ ANDs│◄────┘        │              │   │
       │    └───┬─┘              │              │   │
    ┌──▼─┐  ┌───▼─┐Cout┌─────────▼─┐    Cout┌───▼───▼───┐
    │ OR │◄─┤ AND │◄───┤    HAs    │◄───────┤    FAs    │◄─Cin
    └─┬──┘  └─────┘    └─────┬─────┘        └─────┬─────┘
      │                      ┼                    ┼
      │                      │                    │
      ▼                      ▼                    ▼
     Cout                   Sums                 Sums
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        increment_block_size (int, optional): Size of each composite select adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of unsigned cia. Defaults to "".
        name (str, optional): Name of unsigned cia. Defaults to "u_cia".
    """
    def __init__(self, a: Bus, b: Bus, increment_block_size: int = 4, prefix: str = "", name: str = "u_cia", **kwargs):
        self.N = max(a.N, b.N)
        super().__init__(inputs=[a, b], prefix=prefix, name=name, out_N=self.N+1, **kwargs)

        # Bus sign extension in case buses have different lengths
        self.a.bus_extend(N=self.N, prefix=a.prefix)
        self.b.bus_extend(N=self.N, prefix=b.prefix)

        assert increment_block_size > 0, "Block size of the individual increment groups must be greater than 0."

        # To signify current number of blocks and number of bits that remain to be added into function blocks
        block_n = 0
        N_wires = self.N
        cin = ConstantWireValue0()
        while N_wires != 0:
            propagate_wires = []
            block_size = increment_block_size if N_wires >= increment_block_size else N_wires
            for i in range(block_size):
                if block_n == 0:  # First block contains one chain of full adders (with Cin) directly connected to output wires
                    self.add_component(FullAdder(a=self.a.get_wire((block_n*increment_block_size)+i), b=self.b.get_wire((block_n*increment_block_size)+i), c=cin, prefix=self.prefix+f"_cin_block{block_n}_fa{i}"))
                else:  # Other blocks contain one chain of full adders accepting input Cin of 0 to calculate intermediate sum/carry/propagate signals in parallel and other chain of half adders to calculate final sum bits and finally AND OR logic to propagate carry into the next increment block
                    inter_cin = ConstantWireValue0() if i == 0 else inter_cin
                    self.add_component(FullAdderP(a=self.a.get_wire((block_n*increment_block_size)+i), b=self.b.get_wire((block_n*increment_block_size)+i), c=inter_cin, prefix=self.prefix+f"_block{block_n}_fa_p{i}"))
                    propagate_wires.append(self.get_previous_component().get_propagate_wire())
                    inter_cin = self.get_previous_component().get_carry_wire()

                    # Chain of HA to obtain final sum bits
                    self.add_component(HalfAdder(a=self.get_previous_component().get_sum_wire(), b=cin, prefix=self.prefix+f"_block{block_n}_ha{i}"))
                # Updating cin for the the next adder
                cin = self.get_previous_component().get_carry_wire()
                # Connecting adder's output sum bit to its proper position within the described circuit's output bus
                self.out.connect(i+(block_n*increment_block_size), self.get_previous_component().get_sum_wire())

            # ANDing of propagate wires, gate's output serves as an input into another AND gate that is used along with the other input (carry from the current increment block) to determine the propagate signal value; next the propagate signal with the carry output from the intermediate FA carry chain is driven to an OR gate to obtain the carry bit for the next increment block
            if block_n != 0:
                propagation_and = MultipleInputLogicGate(a=Bus(prefix=self.prefix+f"_propagate_signal{block_n}", N=len(propagate_wires), wires_list=propagate_wires), two_input_gate_cls=AndGate, parent_component=self, prefix=self.prefix+f"_and_propagate{block_n}")
                self.add_component(AndGate(a=propagation_and.out, b=cin, prefix=self.prefix+"_and"+str(self.get_instance_num(cls=AndGate)), parent_component=self))
                self.add_component(OrGate(a=self.get_previous_component().out, b=inter_cin, prefix=self.prefix+"_or"+str(self.get_instance_num(cls=OrGate)), parent_component=self))
                cin = self.get_previous_component().out

            N_wires -= block_size
            block_n += 1
        # Connection of final Cout
        self.out.connect(self.N, cin)


class SignedCarryIncrementAdder(UnsignedCarryIncrementAdder, GeneralCircuit):
    """Class representing signed carry increment adder.

    Carry increment adder represents a modified carry select adder that achieves about the same critical delay
    while containing less circuitry. It achieves this by factoring out the common logic from the two
    chains of full adders (one with cin of 0, the other with 1) present in a classical carry select adder and
    by replacing multiplexers with HAs.

    Each carry increment block is composed of these logic parts:
    Full adder chain with cin 0 is used to obtain intermediate sum, propagate and cout signals. (In case of first block actual sum bits are obtained)
    From second block onward the blocks also contain a chain of half adders to calculate the final sum bits.
    Finally the intermediate propagate signals are ANDed with the cout of the half adder chain and ORed with the intermediate
    full adder cout to determine the carry bit for the next increment block.
    At last XOR gates are used to ensure proper sign extension.

    ```
                            ┼   ┼                ┼   ┼
                    Cout┌───▼───▼───┐            │   │
        ┌───────────────┤    FAs    │◄─0         │   │
        │               └┬────────┬─┘            │   │
        │    ┌─────┐     ┼ P sigs ┼ Sums         │   │
        │    │ ANDs│◄────┘        │              │   │
        │    └───┬─┘              │              │   │
     ┌──▼─┐  ┌───▼─┐Cout┌─────────▼─┐    Cout┌───▼───▼───┐
     │ OR │◄─┤ AND │◄───┤    HAs    │◄───────┤    FAs    │◄─Cin
     └─┬──┘  └─────┘    └─────┬─────┘        └─────┬─────┘
    ┌──▼───┐                  ┼                    ┼
    │ SIGN │                  │                    │
    │Extend│                  │                    │
    └──┬───┘                  │                    │
       │                      │                    │
       │                      │                    │
       ▼                      ▼                    ▼
      Cout                   Sums                 Sums
    ```

    Description of the __init__ method.

    Args:
        a (Bus): First input bus.
        b (Bus): Second input bus.
        increment_block_size (int, optional): Size of each composite select adder block size. Defaults to 4.
        prefix (str, optional): Prefix name of signed cia. Defaults to "".
        name (str, optional): Name of signed cia. Defaults to "s_cia".
    """
    def __init__(self, a: Bus, b: Bus, increment_block_size: int = 4, prefix: str = "", name: str = "s_cia", **kwargs):
        # Signed bus length extension
        N = max(a.N, b.N)
        a.bus_extend(N=N, prefix=a.prefix, desired_extension_wire=a.get_wire(a.N-1))
        b.bus_extend(N=N, prefix=b.prefix, desired_extension_wire=b.get_wire(b.N-1))
        
        super().__init__(a=a, b=b, increment_block_size=increment_block_size, prefix=prefix, name=name, signed=True, **kwargs)

        # Additional XOR gates to ensure correct sign extension in case of sign addition
        self.add_component(XorGate(self.a.get_wire(self.N-1), self.b.get_wire(self.N-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self))
        self.add_component(XorGate(self.get_previous_component().out, self.out.get_wire(-1), prefix=self.prefix+"_xor"+str(self.get_instance_num(cls=XorGate, count_disabled_gates=False)), parent_component=self))
        self.out.connect(self.N, self.get_previous_component().out)



from .two_input_one_bit_components import (
    HalfAdder,
    HalfSubtractor,
    PGLogicBlock
)

from .three_input_one_bit_components import (
    FullAdder,
    FullAdderP,
    FullAdderPG,
    PGSumLogic,
    TwoOneMultiplexer,
    FullSubtractor,
    GreyCell
)

from .four_input_one_bit_components import (
    BlackCell
)
